import pygame as pg

import bluesky as bs


white = (255, 255, 255)

class Menu:
    """
    Menu class definition : Menu button window

    # To adapt menu content:
    #
    # 1. Edit screen in powerpoint file in ./devtools/menu
    # 2. Save all slide as jpg
    # 3. In the resulting subfolder run rename.bat,
    #    which renames and copies files to right location
    # 4. Add commands in menu.dat
    #

    Methods:
        Menu()         :  constructor

    Created by  : Jacco M. Hoekstra (TU Delft)
    """
    def __init__(self,win,x,y):
        self.win   = win    # Screen to draw window on
        self.x     = x      # x-position
        self.y     = y      # y-position
        self.ipage = 0      # which menu
        self.npages = 4     # Number of pages

        # Read, crop and scale powerpoint bitmaps
        self.bmps  = []     # list with menu bitmpas

        for i in range(self.npages):
            imgpath = bs.resource("graphics/menu/menu"+str(i+1)+".jpg")

            im = pg.image.load(imgpath)

            image = im.subsurface(pg.Rect(200, 194, 222, 330))
            width, height = image.get_size() # Get dimensions
            surface = pg.transform.smoothscale(image,\
                                               (int(width/2),int(height/2)))

            rect = surface.get_rect()
            pg.draw.rect(surface, white, rect,1)
            self.bmps.append(surface)

        # Get dimensions
        self.dx,self.dy = surface.get_size()
        self.rect = pg.Rect(self.x,self.y,self.dx,self.dy)

        # Read data
        f = open(bs.resource("graphics/menu.dat"))
        self.cmd = []
        ipage = 0
        page = []
        for line in f.readlines():

            if len(line.strip())==0:
                continue
            elif line.lower().strip()[:5]=="start": # starting page number
                self.ipage = int(line.split("=")[1])
            elif line[0].strip()==">":
                ipage = ipage+1
                if len(page)<20:
                    page=page+(20-len(page))*[""]
                self.cmd.append(page)
                page = []
            else:
                page.append(line[:-1])
        f.close()

        # Clean up
        del self.cmd[0]  # Remove first empty page due to >Traffic line
        self.ipage = max(0,min(self.npages,self.ipage)) # limit ipage

        return

    def update(self):

        self.rect = pg.Rect(self.x,self.y,self.dx,self.dy)

        return self.bmps[self.ipage]

    def getcmd(self,mpos):
        ix  = max(0,min(1,int((mpos[0]-self.x)/54)))
        idx = max(0,int((mpos[1]-self.y)/15))
        if ix == 0:
            self.ipage = min(self.npages-1,int(idx/2))
            cmdtxt =""
        else:
            cmdtxt = self.cmd[self.ipage][min(20,idx)]

        return cmdtxt



import pickle
import bluesky as bs

## Default settings
bs.settings.set_variable_defaults(cache_path='cache')


def openfile(*args):
    return CacheFile(*args)


class CacheError(Exception):
    ''' Exception class for CacheFile errors. '''
    pass


class CacheFile():
    ''' Convenience class for loading and saving pickle cache files. '''
    def __init__(self, fname, version_ref='1'):
        self.fname = bs.resource(bs.settings.cache_path).joinpath(fname)
        self.version_ref = version_ref
        self.file = None

    def check_cache(self):
        ''' Check whether the cachefile exists, and is of the correct version. '''
        if not self.fname.is_file():
            raise CacheError('Cachefile not found: ' + str(self.fname))

        self.file = open(self.fname, 'rb')
        version = pickle.load(self.file)

        # Version check
        if not version == self.version_ref:
            self.file.close()
            self.file = None
            raise CacheError('Cache file out of date: ' + str(self.fname))
        print('Reading cache:', self.fname)

    def load(self):
        ''' Load a variable from the cache file. '''
        if self.file is None:
            self.check_cache()

        return pickle.load(self.file)

    def dump(self, var):
        ''' Dump a variable to the cache file. '''
        if self.file is None:
            self.file = open(self.fname, 'wb')
            pickle.dump(self.version_ref, self.file, pickle.HIGHEST_PROTOCOL)
            print("Writing cache:", self.fname)
        pickle.dump(var, self.file, pickle.HIGHEST_PROTOCOL)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.file:
            self.file.close()



import numpy as np
import bluesky as bs
from bluesky.tools.aero import nm, g0
from bluesky.tools.misc import degto180
from bluesky.core import Entity


class ActiveWaypoint(Entity, replaceable=True):
    def __init__(self):
        super().__init__()
        with self.settrafarrays():
            self.lat         = np.array([])    # [deg] Active WP latitude
            self.lon         = np.array([])    # [deg] Active WP longitude
            self.nextturnlat = np.array([])    # [deg] Next turn WP latitude
            self.nextturnlon = np.array([])    # [deg] Next turn WP longitude
            self.nextturnspd = np.array([])    # [m/s] Next turn WP speed
            self.nextturnbank= np.array([])    # [deg] Next turn WP bank angle
            self.nextturnrad = np.array([])    # [m]   Next turn WP turn radius
            self.nextturnhdgr= np.array([])  # [deg/s] Next turn WP heading rate (<0 => not specified)
            self.nextturnidx = np.array([])    # [-]   Next turn WP index
            self.nextaltco   = np.array([])    # [m] Altitude to arrive at after distance xtoalt
            self.xtoalt      = np.array([])    # [m] Distance to next altitude constraint
            self.nextspd     = np.array([])    # [CAS[m/s]/Mach] save speed from next wp for next leg
            self.spd         = np.array([])    # [CAS[m/s]/Mach]Active WP speed (constraint or calculated)
            self.spdcon      = np.array([])    # [CAS[m/s]/Mach]Active WP speed constraint
            self.vs          = np.array([])    # [m/s] Active vertical speed to use
            self.turndist    = np.array([])    # [m] Distance when to turn to next waypoint
            self.flyby       = np.array([])    # Flyby switch, when False, flyover (turndist=0.0)
            self.flyturn     = np.array([])    # Flyturn switch, customised turn parameters; when False, use flyby/flyover
            self.turnbank    = np.array([])    # Flyturn turn bank angle [deg]
            self.turnrad     = np.array([])    # Flyturn turn radius (<0 => not specified)
            self.turnspd     = np.array([])    # [m/s, CAS] Flyturn turn speed for next turn (<=0 => not specified)
            self.turnhdgr    = np.array([])    # [deg/s]Flyturn turn heading rate (<0 => not specified)
            self.oldturnspd  = np.array([])    # [TAS, m/s] Flyturn turn speed for previous turn (<=0 => not specified)
            self.turnfromlastwp = np.array([]) # Currently in flyturn-mode from last waypoint (old turn, beginning of leg)
            self.turntonextwp =  np.array([])  # Currently in flyturn-mode to next waypoint (new flyturn mode, end of leg)
            self.torta       = np.array([])    # [s] Next req Time of Arrival (RTA) (-999. = None)
            self.xtorta      = np.array([])    # [m] distance to next RTA
            self.next_qdr    = np.array([])    # [deg] track angle of next leg
            self.swlastwp    = np.array([],dtype=bool) # switch indicating this is the last waypoint
            self.curlegdir   = np.array([])    # [deg] direction to active waypoint upon activation
            self.curleglen   = np.array([])    # [deg] direction to active waypoint upon activation

    def create(self, n=1):
        super().create(n)
        # LNAV route navigation
        self.lat[-n:]        = 0.       # [deg]Active WP latitude
        self.lon[-n:]        = 0.       # [deg]Active WP longitude
        self.nextturnlat[-n:]= 0        # [deg] Next turn WP latitude
        self.nextturnlon[-n:]= 0        # [deg] Next turn WP longitude
        self.nextturnspd[-n:]= -999.    # [m/s] Next turn WP speed
        self.nextturnbank[-n:] = -999.  # [deg] Next turb WP bank angle
        self.nextturnrad[-n:]= -999.    # [m]   Next turn WP radius
        self.nextturnhdgr[-n:]= -999.   # [deg/s] Next turn WP heading rate (<0 => not specified)
        self.nextturnidx[-n:]= -999.    # [-] Next turn WP index
        self.nextaltco[-n]   = -999.    # [m] Altitude to arrive at after distance xtoalt
        self.xtoalt[-n:]     = 0.0      # [m] Distance to next altitude constraint
        self.nextspd[-n:]    = -999.    # [CAS[m/s]/Mach]Next leg speed from current WP
        self.spd[-n:]        = -999.    # [CAS[m/s]/Mach]Active WP speed
        self.spdcon[-n:]     = -999.    # [CAS[m/s]/Mach]Active WP speed constraint
        self.turndist[-n:]   = 1.0      # [m] Distance to active waypoint where to turn
        self.flyby[-n:]      = 1.0      # Flyby/fly-over switch
        self.flyturn[-n:]    = False    # Flyturn switch, when False, when False, use flyby/flyover
        self.turnbank[-n:]   = -999.    # Flyturn turn bank angle [deg]
        self.turnrad[-n:]    = -999.    # [m] Flyturn turn radius (<0 => not specified)
        self.turnspd[-n:]    = -999.    # [m/s]Flyturn turn speed (<0 => not specified)
        self.turnhdgr[-n:]   = -999.    # [deg/s]Flyturn turn heading rate (<0 => not specified)
        self.oldturnspd[-n:] = -999.    # [TAS, m/s] Flyturn turn speed for previous turn (<=0 => not specified)
        self.turnfromlastwp[-n:] = False # Currently in flyturn-mode from last waypoint (old turn, beginning of leg)
        self.turntonextwp[-n:] = False  # Currently in flyturn-mode to next waypoint (new flyturn mode, end of leg)
        self.torta[-n:]      = -999.0   # [s] Req Time of Arrival (RTA) for next wp (-999. = None)
        self.xtorta[-n:]     = 0.0      # Distance to next RTA
        self.next_qdr[-n:]   = -999.0   # [deg] bearing next leg
        self.swlastwp[-n:]   = False    # Switch indicating active waypoint is last waypoint
        self.curlegdir[-n:]  = -999.0   # [deg] direction to active waypoint upon activation
        self.curleglen[-n:]  = -999.0   # [nm] distance to active waypoint upon activation
  
    def reached(self, qdr, dist):
        # Calculate distance before waypoint where to start the turn
        # Note: this is a vectorized function, called with numpy traffic arrays
        # It returns the indices where the Reached criterion is True
        #
        # Turn radius:      R = V2 tan phi / g
        # Distance to turn: wpturn = R * tan (1/2 delhdg) but max 4 times radius
        # using default bank angle per flight phase
        # Gather required data
        flyby = self.flyby
        flyturn = self.flyturn

        # Turn dist is zero for flyover, and it is previously calculated in autopilot for others
        self.turndist = np.logical_or(flyby,flyturn)*self.turndist

        # Avoid circling by checking too close to waypoint based on ground speed, assumption using vicinity criterion:
        # flying away and within 4 sec distance based on ground speed (4 sec = sensitivity tuning parameter)

        close2wp = dist/(np.maximum(0.0001,np.abs(bs.traf.gs)))<4.0 # Waypoint is within 4 seconds flight time
        tooclose2turn = close2wp*(np.abs(degto180(bs.traf.trk % 360. - qdr % 360.)) > 90.)

        # When too close to waypoint or we have passed the active waypoint, based on leg direction,switch active waypoint
        # was:  away  = np.logical_or(close2wp,swlastwp)*(np.abs(degto180(bs.traf.trk%360. - qdr%360.)) > 90.) # difference large than 90
        awayorpassed =  np.logical_or(tooclose2turn,np.abs(degto180(qdr-bs.traf.actwp.curlegdir))>90.)

        # Should no longer be needed with leg direction
        # Ratio between distance close enough to switch to next wp when flying away
        # When within pro1 nm and flying away: switch also
        #proxfact = 1.02 # Turnradius scales this contant , factor => [turnrad]
        #incircle = dist<turnrad*proxfact
        #circling = away*incircle # [True/False] passed wp,used for flyover as well

        # Check whether shift based dist is required, set closer than WP turn distance
        # Detect indices
        #swreached = np.where(bs.traf.swlnav * np.logical_or(awayorpassed,np.logical_or(dist < self.turndist,circling)))[0]
        swreached = np.where(bs.traf.swlnav * np.logical_or(awayorpassed,dist < self.turndist))[0]

        # Return indices for which condition is True/1.0 for a/c where we have reached waypoint
        return swreached

    # Calculate turn distance for scalars
    def calcturn(self, acidx, tas , wpqdr, next_wpqdr, turnbank, turnrad, turnspd, turnhdgr, flyturn, flyby):
        """Calculate the properties of a turn in function of the input. Inputs are SCALARS."""
        # If this is not a flyturn waypoint, just return standard values
        # Calculate the sum of bools to figure out things more easily
        num_defined_values = sum([turnbank>0, turnrad>0, turnspd>0, turnhdgr>0])
        # Case 1: No value is given or not a flyturn, use defaults
        if num_defined_values == 0 or not flyturn:
            turnbank = np.rad2deg(bs.traf.ap.bankdef[acidx])
            turnspd = tas
            # Calculate the radius
            turnrad = turnspd**2/(g0*np.tan(np.deg2rad(turnbank)))
            if not flyby:
                # This is a flyover waypoint, so we have to fly OVER it, so turndist is 0
                return 0, turnrad, turnspd, turnbank, turnhdgr
        # Case 2: only one value is given. We then want to keep the speed as TAS unless the speed is the one that isn't specified
        elif num_defined_values == 1:
            # If the velocity is the given one, take default bank and calculate remaining values
            if turnspd>0:
                turnbank = 25
                turnrad = turnspd**2/(g0*np.tan(np.deg2rad(turnbank)))
            #Otherwise, keep velocity constant and calculate the remaining values
            elif turnrad>0:
                turnspd = tas
                turnbank = np.arctan(turnspd**2/(turnrad * g0))
                
            elif turnbank>0:
                turnspd = tas
                turnrad = turnspd**2 / (g0 * np.tan(np.deg2rad(turnbank)))
                
            elif turnhdgr>0:
                turnspd = tas
                #Calculate the other two
                turnbank = np.arctan(turnhdgr*turnspd/g0)
                turnrad = turnspd / turnhdgr
        # Case 3: We have two defined values and need to calculate the other
        elif num_defined_values == 2:
            # Need to figure out which ones and calculate the remaining
            if turnrad>0 and turnbank>0:
                turnspd = np.sqrt(g0*turnrad*np.tan(np.deg2rad(turnbank)))
            elif turnrad>0 and turnspd>0:
                turnbank = np.arctan(turnspd**2/(turnrad*g0))
            elif turnspd>0 and turnbank>0:
                turnrad = turnspd**2/(g0*np.tan(np.deg2rad(turnbank)))
            # Now the cases where one of the known vars is heading rate
            elif turnhdgr>0 and turnspd>0:
                turnbank = np.arctan(turnhdgr*turnspd/g0)
                turnrad = turnspd / turnhdgr
            elif turnhdgr>0 and turnbank>0:
                turnspd = g0*np.tan(np.deg2rad(turnbank))/turnhdgr
                turnrad = turnspd / turnhdgr
            elif turnhdgr>0 and turnrad>0:
                turnspd = turnhdgr * turnrad
                turnbank = np.arctan(turnhdgr*turnspd/g0)
        
        # turndist is in meters
        turndist = np.abs(turnrad * np.tan(np.radians(0.5 * np.abs(degto180(wpqdr%360. - next_wpqdr%360.)))))
        return turndist, turnrad, turnspd, turnbank, turnhdgr



"""
Airspace Design Contest Scenario Generator:

Draw circles and runways in use
Generate traffic at edges and at airports
Make sure scenario can be saved

"""


from bluesky import stack,traf,sim,tools,navdb  #, settings, navdb, traf, sim, scr, tools
from bluesky.plugins.trafgenclasses import Source, Drain, setcircle
from bluesky.tools.position import txt2pos
from bluesky.tools.geo import kwikpos
#from bluesky.tools import areafilter
#from bluesky.tools.aero import vtas2cas,ft
#from bluesky.tools.misc import degto180


import numpy as np

# Default values
swcircle   = False # No active circle
ctrlat     = 52.6  # Default values to resolve navdb names
ctrlon     = 5.4
radius     = 230.
globalgain = 1.0


def init_plugin():
    print("Initialising contest scenario generator")

    # Create an empty geovector list
    reset()

    # Configuration parameters
    config = {
        # The name of your plugin
        'plugin_name'      : 'TRAFGEN',
        'plugin_type'      : 'sim',
        'update_interval'  :  0.1,

        # The update function is called after traffic is updated.
        'update':          update,

        # The preupdate function is called before traffic is updated. Use this
        #'preupdate':       preupdate,

        # Reset contest
        'reset':         reset
        }

    # Add two commands: GEOVECTOR to define a geovector for an area
    stackfunctions = {
        # Starting contest traffic generation
        'TRAFGEN': [
            'TRAFGEN [location],cmd,[arg, arg, ...]',
            'string',
            trafgencmd,
            'CONTEST command']
        }

    # init_plugin() should always return these two dicts.

    return config, stackfunctions

def reset():
    # Contest global variables
    global ctrlat,ctrlon,radius,dtsegment,drains,sources,rwsdep,rwsarr

    # Set default parameters for spawning circle

    swcircle = False
    ctrlat = 52.6  # [deg]
    ctrlon = 5.4  # [deg]
    radius = 230.0  # [nm]

    # Draw circle
    stack.stack("CIRCLE SPAWN," + str(ctrlat) + "," + str(ctrlon) + "," + str(radius))

    # Average generation interval in [s] per segment
    dtsegment = 12 * [1.0]

    # drains: dictionary of drains
    sources     = dict([])
    drains      = dict([])

    return


### Periodic update functions that are called by the simulation. You can replace
### this by anything, so long as you communicate this in init_plugin

#def preupdate(): # To be safe preupdate is used iso update
#    pass
#    return


def update(): # Update all sources and drain
    for src in sources:
        sources[src].update(globalgain)

    for drn in drains:
        drains[drn].update(globalgain)

    return

### Other functions of your plug-in
def trafgencmd(cmdline):
    global ctrlat,ctrlon,radius,dtsegment,drains,sources,rwsdep,rwsarr,globalgain

    cmd,args = splitline(cmdline)

    #print("TRAFGEN: cmd,args=",cmd,args)

    if cmd=="CIRCLE" or cmd=="CIRC":

        # Draw circle
        try:
            swcircle = True
            ctrlat = float(args[0])
            ctrlon = float(args[1])
            radius = float(args[2])
            setcircle(ctrlat, ctrlon, radius)
        except:
            return False,'TRAFGEN ERROR while reading CIRCLE command arguments (lat,lon,radius):'+str(args)
        stack.stack("DEL SPAWN")
        stack.stack("CIRCLE SPAWN," + str(ctrlat) + "," + str(ctrlon) + "," + str(radius))

    elif cmd=="GAIN" or cmd=="FACTOR":
        # Set global gain on traffic density
        try:
            globalgain = float(args[0])
        except:
            return False, "TRAFGEN GAIN error: invalid value "+args[0]


    elif cmd=="SRC" or cmd == "SOURCE": # Define streams by source, give destinations
        name = args[0].upper()
        cmd = args[1].upper()
        cmdargs = args[2:]
        if name not in sources:
            if not name[:4] == "SEGM":
                success,posobj = txt2pos(name,ctrlat,ctrlon)
                if success:
                    aptlat, aptlon = posobj.lat, posobj.lon
                    sources[name] = Source(name,cmd,cmdargs)
                    if posobj.type == "rwy" and name.count("/")==1:
                        aptname, rwyname = name.split('/')
                        sources[name].polys += drawrwy(aptname,[rwyname],posobj.lat,posobj.lon,drawdeprwy)

            else:
                try:
                    brg = int(name[4:])
                    aptlat,aptlon = kwikpos(ctrlat,ctrlon,brg,radius)
                    sources[name] = Source(name, cmd, cmdargs)
                    success = True
                except:
                    success = False
        else:
            aptlat,aptlon = sources[name].lat,sources[name].lat
            success = True

        if success:
            if cmd[:6]=="RUNWAY" or cmd=="RWY":
                sources[name].setrunways(cmdargs)
                for polyname in sources[name].polys:
                    stack.stack("DEL "+polyname)
                sources[name].polys += drawrwy(name,cmdargs,aptlat,aptlon,drawdeprwy)

            elif cmd=="DEST":
                success = sources[name].adddest(cmdargs)
            elif cmd=="FLOW":
                success = sources[name].setflow(cmdargs[0])
            elif cmd=="TYPES" or cmd=="TYPE":
                sources[name].addactypes(cmdargs)
            elif cmd=="ALT": # Set fixed alt or interval
                sources[name].setalt(cmdargs)
            elif cmd=="SPD":
                sources[name].setspd(cmdargs)
            elif cmd=="HDG":
                sources[name].sethdg(cmdargs)
            if not success:
                return False, "TRAFGEN SRC ERROR"+cmd+" ".join(cmdargs)
        else:
            return False,"TRAFGEN SRC ERROR "+name+" NOT FOUND"

    elif cmd=="DRN" or cmd=="DRAIN":
        name = args[0].upper()
        cmd = args[1].upper()
        cmdargs = args[2:]
        if name in drains:
            aptlat, aptlon = drains[name].lat, drains[name].lat
            success = True
        else:
            # name not in drains: New drain defined
            if not name[:4]=="SEGM":
                success, posobj = txt2pos(name, ctrlat, ctrlon)
                if success:
                    drains[name] = Drain(name, cmd, cmdargs)

                    if posobj.type == "rwy" and name.count("/") == 1:
                        aptname, rwyname = name.split('/')
                        drains[name].polys += drawrwy(aptname, [rwyname], posobj.lat, posobj.lon, drawapprwy)


            else:
                try:
                    brg = int(name[:4])
                    drains[name] = Drain(name, cmd, cmdargs)
                    success = True
                except:
                    success = False


        if success:
            if cmd[:6] == "RUNWAY" or cmd == "RWY":
                # Delete old runways
                for polyname in drains[name].polys:
                    stack.stack("DEL "+polyname)
                aptlat, aptlon = drains[name].lat,drains[name].lon
                drains[name].setrunways(cmdargs)
                drains[name].polys += drawrwy(name,cmdargs,aptlat,aptlon,drawapprwy)

            elif cmd=="ORIG":
                success = drains[name].addorig(cmdargs)

            elif cmd=="FLOW":
                sucess = drains[name].setflow(cmdargs[0])
            elif cmd=="TYPES" or cmd=="TYPE":
                drains[name].addactypes(cmdargs)
            elif cmd == "ALT":  # Set fixed alt or interval
                drains[name].setalt(cmdargs)
            elif cmd == "SPD":
                drains[name].setspd(cmdargs)
            elif cmd == "HDG":
                drains[name].sethdg(cmdargs)
            if not success:
                return False, "TRAFGEN SRC ERROR"+cmd+" ".join(cmdargs)

        else:
            return False, "TRAFGEN DRN ERROR " + name + " NOT FOUND"

    return True

def splitline(rawline):
    # Interpet string like a command with arguments
    # Replace multiple spaces by one space

    line = rawline.strip().upper()
    if line.count("#")>=1:
        icomment = line.index("#")
        line  = line[:icomment]

    while line.count("  ") > 0:
        line = line.replace("  ", " ")

    # and remove spaces around commas
    while line.count(", ") > 0:
        line = line.replace(", ", ",")

    while line.count(" ,") > 0:
        line = line.replace(" ,", ",")

    # Replace remaining spaces, which are separators, by comma
    line = line.strip().replace(" ",",")

    # Split using commas
    args = line.split(",")
    if len(args)>=1:
        cmd = args[0]
    else:
        cmd = ""
    return cmd,args[1:]

def drawrwy(aptname,cmdargs,aptlat,aptlon,drawfunction):
    rwnames = []

    for rwy in cmdargs:
        if rwy[0] == "R":
            success, rwyposobj = txt2pos(aptname + "/" + rwy, aptlat, aptlon)
        else:
            success, rwyposobj = txt2pos(aptname + "/RW" + rwy, aptlat, aptlon)

        if not success:
            success,rwyposobj = txt2pos(aptname,aptlat,aptlon)

        if success:
            rwydigits = rwy.lstrip("RWY").lstrip("RW")

            # Look up runwayhdg
            try:
                rwyhdg = navdb.rwythresholds[aptname][rwydigits][2]
            except:
                try:
                    rwyhdg = navdb.rwythresholds[aptname][rwydigits.lstrip("0")][2]
                except:
                    rwyhdg = 10.*int(rwydigits.rstrip("LCR").lstrip("0"))

            rwnames.append(drawfunction(aptname, rwy, rwyposobj.lat, rwyposobj.lon, rwyhdg))
        else:
            # Use airport lat,lon en given heading
            stack.stack("ECHO TARFGEN RUNWAY "+aptname+"/"+rwy+" NOT FOUND")

    return rwnames


def drawapprwy(apt,rwy,rwylat,rwylon,rwyhdg):
    # Draw approach ILS arrow
    Lapp  =  7. # [nm] length of approach path drawn
    phi   =  5. # [deg] angle of half the arrow

    # Calculate arrow (T = Threshold runway):
    #                               /------------- L   (left)
    #                 /-----------/              /
    #   T -------------------------------------- A    (approach)
    #                 \-----------\              \
    #                              \--------------R    (right)
    #

    applat,applon     = kwikpos(rwylat,rwylon,(rwyhdg+180)%360.,Lapp)
    rightlat,rightlon = kwikpos(rwylat,rwylon,(rwyhdg+180-phi)%360,Lapp*1.1)
    leftlat,leftlon   = kwikpos(rwylat,rwylon,(rwyhdg+180+phi)%360,Lapp*1.1)

    # Make arguments for POLYLINE command
    T = str(rwylat)+","+str(rwylon)
    A = str(applat)+","+str(applon)
    L = str(leftlat)+","+str(leftlon)
    R = str(rightlat)+","+str(rightlon)

    stack.stack("POLYLINE "+apt+rwy+"-A,"+",".join([T,A,L,T,R,A]))

    return apt+rwy+"-A"

def drawdeprwy(apt,rwy,rwylat,rwylon,rwyhdg):
    # Draw approach ILS arrow
    Ldep  =  5. # [nm] length of approach path drawn
    phi   =  3. # [deg] angle of half the arrow

    # Calculate arrow (T = Threshold runway):
    #      L   (left)
    #     /
    #   D ------------------------------------- T    (approach)
    #     \
    #      R    (right)
    #

    deplat,deplon     = kwikpos(rwylat,rwylon,rwyhdg%360.,Ldep*1.1)
    rightlat,rightlon = kwikpos(rwylat,rwylon,(rwyhdg+phi)%360,Ldep)
    leftlat,leftlon   = kwikpos(rwylat,rwylon,(rwyhdg-phi)%360,Ldep)

    # Make arguments for POLYLINE command
    T = str(rwylat)+","+str(rwylon)
    D = str(deplat)+","+str(deplon)
    L = str(leftlat)+","+str(leftlon)
    R = str(rightlat)+","+str(rightlon)

    stack.stack("POLYLINE " + apt + rwy + "-D," + ",".join([R,D,L,D,T]))
   

    return apt + rwy + "-D"



""" Pilot logic."""
import numpy as np
import bluesky as bs
from bluesky.core import TrafficArrays

class APorASAS(TrafficArrays):
    def __init__(self):
        super().__init__()
        with self.settrafarrays():
            # Desired aircraft states
            self.alt = np.array([])  # desired altitude [m]
            self.hdg = np.array([])  # desired heading [deg]
            self.trk = np.array([])  # desired track angle [deg]
            self.vs  = np.array([])  # desired vertical speed [m/s]
            self.tas = np.array([])  # desired speed [m/s]

    def create(self, n=1):
        super().create(n)
        self.alt[-n:] = bs.traf.alt[-n:]
        self.tas[-n:] = bs.traf.tas[-n:]
        self.hdg[-n:] = bs.traf.hdg[-n:]
        self.trk[-n:] = bs.traf.trk[-n:]

    def update(self):
        #--------- Input to Autopilot settings to follow: destination or ASAS ----------
        # Convert the ASAS commanded speed from ground speed to TAS
        if bs.traf.wind.winddim > 0:
            vwn, vwe     = bs.traf.wind.getdata(bs.traf.lat, bs.traf.lon, bs.traf.alt)
            asastasnorth = bs.traf.cr.tas * np.cos(np.radians(bs.traf.cr.trk)) - vwn
            asastaseast  = bs.traf.cr.tas * np.sin(np.radians(bs.traf.cr.trk)) - vwe
            asastas      = np.sqrt(asastasnorth**2 + asastaseast**2)
        # no wind, then ground speed = TAS
        else:
            asastas = bs.traf.cr.tas # TAS [m/s]

        # Select asas if there is a conflict AND resolution is on
        # Determine desired states per channel whether to use value from ASAS or AP.
        # bs.traf.cr.active may be used as well, will set all of these channels
        self.trk = np.where(bs.traf.cr.hdgactive, bs.traf.cr.trk, bs.traf.ap.trk)
        self.tas = np.where(bs.traf.cr.tasactive, asastas, bs.traf.ap.tas)
        self.alt = np.where(bs.traf.cr.altactive, bs.traf.cr.alt, bs.traf.ap.alt)
        self.vs  = np.where(bs.traf.cr.vsactive, bs.traf.cr.vs, bs.traf.ap.vs)

        # ASAS can give positive and negative VS, but the sign of VS is determined using delalt in Traf.ComputeAirSpeed
        # Therefore, ensure that pilot.vs is always positive to prevent opposite signs of delalt and VS in Traf.ComputeAirSpeed
        self.vs = np.abs(self.vs)

        # Compute the desired heading needed to compensate for the wind
        if bs.traf.wind.winddim > 0:

            # Calculate wind correction
            vwn, vwe = bs.traf.wind.getdata(bs.traf.lat, bs.traf.lon, bs.traf.alt)
            Vw       = np.sqrt(vwn * vwn + vwe * vwe)
            winddir  = np.arctan2(vwe, vwn)
            drift    = np.radians(self.trk) - winddir  # [rad]
            steer    = np.arcsin(np.minimum(1.0, np.maximum(-1.0,
                                     Vw * np.sin(drift) / np.maximum(0.001, bs.traf.tas))))
            # desired heading
            self.hdg = (self.trk + np.degrees(steer)) % 360.
        else:
            self.hdg = self.trk % 360.




""" Route implementation for the BlueSky FMS."""
import math
from weakref import WeakValueDictionary
import numpy as np
import bluesky as bs
from bluesky.tools import geo
from bluesky.core import Base
from bluesky.tools.aero import ft, kts, g0, nm, mach2cas, casormach2tas
from bluesky.tools.misc import degto180, txt2tim, txt2alt, txt2spd
from bluesky.tools.position import txt2pos
from bluesky import stack
from bluesky.stack.cmdparser import Command, command, commandgroup



class Route(Base):
    """
    Route class definition   : Route data for an aircraft
    (basic FMS functionality)

    addwpt(name,wptype,lat,lon,alt) :
    Add waypoint (closest to lat/lon when from navdb

    For lat/lon waypoints: use call sign as wpname, number will be added

    Created by  : Jacco M. Hoekstra
    """

    # Waypoint types:
    wplatlon = 0   # lat/lon waypoint
    wpnav    = 1   # VOR/nav database waypoint
    orig     = 2   # Origin airport
    dest     = 3   # Destination airport
    calcwp   = 4   # Calculated waypoint (T/C, T/D, A/C)
    runway   = 5   # Runway: Copy name and positions

    # Aircraft route objects
    _routes: WeakValueDictionary[str, 'Route'] = WeakValueDictionary()

    def __init__(self, acid):
        super().__init__()
        # Add self to dictionary of all aircraft routes
        Route._routes[acid] = self
        # Aircraft id (callsign) of the aircraft to which this route belongs
        self.acid = acid
        self.nwp = 0

       # Waypoint data
        self.wpname = []    # List of waypoint names for this flight plan
        self.wptype = []    # List of waypoint types
        self.wplat  = []    # List of waypoint latitudes
        self.wplon  = []    # List of waypoint longitudes
        self.wpalt  = []    # [m] negative value means not specified
        self.wpspd  = []    # [m/s] negative value means not specified
        self.wprta  = []    # [m/s] negative value means not specified
        self.wpflyby = []   # Flyby (True)/flyover(False) switch
        self.wpstack = []   # Stack with command execured when passing this waypoint

        # Made for drones: fly turn mode, means use specified turn radius and optionally turn speed
        self.wpflyturn  = []   # Flyturn (True) or flyover/flyby (False) switch
        self.wpturnbank = []   # [deg] Bank angle per waypoint
        self.wpturnrad  = []   # [nm] Turn radius per waypoint (<0 = not specified)
        self.wpturnspd  = []   # [kts] Turn speed (IAS/CAS) per waypoint (<0 = not specified)
        self.wpturnhdgr = [] # [deg/s] Heading rate, uses actual speed to calculate bank & radius (<0 = not specified)

        # Current actual waypoint
        self.iactwp = -1

        # Set to default addwpt wpmode
        # Note that neither flyby nor flyturn means: flyover)
        self.swflyby   = True    # Default waypoints are flyby waypoint
        self.swflyturn = False  # Default waypoints are waypoints w/o specified turn

        # Default turn values to be used in flyturn mode
        self.bank = 25.
        self.turnbank  = -999.   # [deg] Negative value indicating no value has been set
        self.turnrad   = -999. # [m] Negative value indicating no value has been set
        self.turnspd   = -999. # [kts] Dito, in this case bank angle of vehicle will be used with current speed
        self.turnhdgr  = -999. # [deg/s] Dito, in this case bank angle of vehicle will be used with current speed
        # The last two defined turn properties. The list will contain variable names, with the latest
        # Changed being at index 1, and the oldest changed being at index 0.
        # For example, I changed SPD and RAD in that order. The list will look like this:
        # acrte_2_defined = ['acrte.turnspd', 'acrte.turnrad']
        # I then change turnbank. The list will now look like this:
        # acrte_2_defined = ['acrte.turnrad', 'acrte.turnbank']
        self.last_2_defined = []

        # if the aircraft lands on a runway, the aircraft should keep the
        # runway heading
        # default: False
        self.flag_landed_runway = False

        self.wpdirfrom = []  # [deg] direction leg to wp
        self.wpdirto   = []  # [deg] direction leg from wp
        self.wpdistto  = []  # [nm] leg length to wp
        self.wpialt    = []
        self.wptoalt   = []  # [m] next alt contraint
        self.wpxtoalt  = []  # [m] distance ot next alt constraint
        self.wptorta   = []  # [s] next time constraint
        self.wpxtorta  = []  # [m] distance to next time constaint

    @staticmethod
    def get_available_name(data, name_, len_=2):
        """
        Check if name already exists, if so add integer 01, 02, 03 etc.
        """
        appi = 0  # appended integer to name starts at zero (=nothing)
        # Use Python 3 formatting syntax: "{:03d}".format(7) => "007"
        fmt_ = "{:0" + str(len_) + "d}"

        # Avoid using call sign without number
        if bs.traf.id.count(name_) > 0:
            appi = 1
            name_ = name_+fmt_.format(appi)

        while data.count(name_) > 0 :
            appi += 1
            name_ = name_[:-len_]+fmt_.format(appi)
        return name_
    
    @stack.command(name = 'ADDWPTMODE')
    @staticmethod
    def addwptMode(acidx: 'acid', mode: 'txt' = None, value: float = None):
        ''' Changes the mode of the ADDWPT command to add waypoints of type 'mode'.
            The mode specified will be used for all ADDWPT commands that are called
            after this command.
            
            Arguments:
            - acid: Aircraft id
            - mode: The mode to use for future waypoints. Options are:
                    FLYBY/FLYOVER/FLYTURN/TURNSPD/TURNRAD/TURNHDG
        '''
        # Get aircraft route
        acid = bs.traf.id[acidx]
        acrte = Route._routes[acid]
        # First, we want to check what 'mode' is, and then call addwptStack 
        # accordingly.
        if mode in ['FLYBY', 'FLYOVER', 'FLYTURN']:
            # We're just changing addwpt mode, call the appropriate function.
            Route.addwptStack(acidx, mode)
            return True
        
        elif mode in  ['TURNSPEED', 'TURNSPD', 'TURNRADIUS', 'TURNRAD', 'TURNBANK', 'TURNPHI', \
                       'TURNHDGRATE', 'TURNHDG','TURNHDGR']:
            # We're changing the turn properties.
            Route.addwptStack(acidx, mode, value)
            return True
            
        elif mode == None:
            # Just echo the current wptmode
            if acrte.swflyby == True and acrte.swflyturn == False:
                bs.scr.echo('Current ADDWPT mode is FLYBY.')
                return True

            elif acrte.swflyby == False and acrte.swflyturn == False:
                bs.scr.echo('Current ADDWPT mode is FLYOVER.')
                return True

            else:
                bs.scr.echo('Current ADDWPT mode is FLYTURN.')
                return True
            
    @stack.command(name='ADDWPT', annotations='acid,wpt,[alt,spd,wpinroute,wpinroute]', aliases=("WPTYPE",))
    @staticmethod
    def addwptStack(acidx, *args):  # args: all arguments of addwpt
        """ADDWPT acid, (wpname/lat,lon),[alt],[spd],[afterwp],[beforewp]"""
        # First get the appropriate ac route
        acid = bs.traf.id[acidx]
        acrte = Route._routes[acid]
        
        #debug print ("addwptStack:",args)
        #print("active = ",self.wpname[self.iactwp])
        #print(argsnwp
        # Check FLYBY or FLYOVER switch, instead of adding a waypoint

        if len(args) == 1:
            swwpmode = args[0].replace('-', '')

            if swwpmode == "FLYBY":
                acrte.swflyby   = True
                acrte.swflyturn = False
                return True

            elif swwpmode == "FLYOVER":
                acrte.swflyby   = False
                acrte.swflyturn = False
                return True

            elif swwpmode == "FLYTURN":
                acrte.swflyby   = False
                acrte.swflyturn = True
                return True

        elif len(args) == 2:

            swwpmode = args[0].replace('-', '')

            if swwpmode == "TURNRAD" or swwpmode == "TURNRADIUS":
                try:
                    if args[1]=="OFF":
                        acrte.turnrad = -999
                        # If this is in last 2 defined, pop it
                        if 'acrte.turnrad' in acrte.last_2_defined:
                            acrte.last_2_defined.pop(acrte.last_2_defined.index('acrte.turnrad'))
                    else:
                        acrte.turnrad = float(args[1]/ft*nm) #arg was originally parsed as wpalt
                        # Well, last we defined is turnrad, so append it to the list
                        acrte.last_2_defined.append('turnrad')
                        # Now, unless the first value is also acrte.turnrad, we set that var to -999 again
                        if len(acrte.last_2_defined) > 2:
                            # Reset first variable if one of the remaining 3
                            if acrte.last_2_defined[0] == 'turnhdgr':
                                acrte.turnhdgr = -999.
                            elif acrte.last_2_defined[0] == 'turnbank':
                                acrte.turnbank = -999.
                            elif acrte.last_2_defined[0] == 'turnspd':
                                acrte.turnspd = -999.
                            # Pop the first one
                            acrte.last_2_defined.pop(0)
                        # Lastly, there is a possibility that we are left with a list of duplicates. Fix it
                        if len(acrte.last_2_defined) == 2 and acrte.last_2_defined[0] == acrte.last_2_defined[1]:
                            # Duplicates, pop the first one
                            acrte.last_2_defined.pop(0)
                        
                except:
                    return False,"Error in processing value of turn radius"

                # Switch flyturn automatically when this is set
                acrte.swflyby = False
                acrte.swflyturn = True

                return True

            elif swwpmode == "TURNSPD" or swwpmode == "TURNSPEED":

                try:
                    if args[1] == "OFF":
                        acrte.turnspd = -999
                    else:
                        acrte.turnspd = args[1]*kts/ft # [m/s] Arg was wpalt Keep it as IAS/CAS orig in kts, now in m/s
                        # Well, last we defined is turnspd, so append it to the list
                        acrte.last_2_defined.append('turnspd')
                        # Now, unless the first value is also acrte.turnspd, we set that var to -999 again
                        if len(acrte.last_2_defined) > 2:
                            # Reset first variable if one of the remaining 3
                            if acrte.last_2_defined[0] == 'turnhdgr':
                                acrte.turnhdgr = -999.
                            elif acrte.last_2_defined[0] == 'turnbank':
                                acrte.turnbank = -999.
                            elif acrte.last_2_defined[0] == 'turnrad':
                                acrte.turnrad = -999.
                            # Pop the first one
                            acrte.last_2_defined.pop(0)
                        # Lastly, there is a possibility that we are left with a list of duplicates. Fix it
                        if len(acrte.last_2_defined) == 2 and acrte.last_2_defined[0] == acrte.last_2_defined[1]:
                            # Duplicates, pop the first one
                            acrte.last_2_defined.pop(0)
                            
                except:
                    return False, "Error in processing value of turn speed"

                # Switch flyturn automatically when this is set
                acrte.swflyby = False
                acrte.swflyturn = True
                
                return True


            elif swwpmode == "TURNHDGRATE" or swwpmode == "TURNHDG" or swwpmode == "TURNHDGR":
                
                try:
                    if args[1] == "OFF":
                        acrte.turnhdgr = -999
                    else:
                        acrte.turnhdgr = args[1]/ft # [deg/s] turn rate
                        # Well, last we defined is turnhdgr, so append it to the list
                        acrte.last_2_defined.append('turnhdgr')
                        # Now, unless the first value is also acrte.turnrad, we set that var to -999 again
                        if len(acrte.last_2_defined) > 2:
                            # Reset first variable if one of the remaining 3
                            if acrte.last_2_defined[0] == 'turnbank':
                                acrte.turnbank = -999.
                            elif acrte.last_2_defined[0] == 'turnspd':
                                acrte.turnspd = -999.
                            elif acrte.last_2_defined[0] == 'turnrad':
                                acrte.turnrad = -999.
                            # Pop the first one
                            acrte.last_2_defined.pop(0)
                        # Lastly, there is a possibility that we are left with a list of duplicates. Fix it
                        if len(acrte.last_2_defined) == 2 and acrte.last_2_defined[0] == acrte.last_2_defined[1]:
                            # Duplicates, pop the first one
                            acrte.last_2_defined.pop(0)
                            
                except:
                    return False, "Error in processing value of turn heading rate"

                # Switch flyturn automatically when this is set
                acrte.swflyby = False
                acrte.swflyturn = True

                return True
            
            elif swwpmode == 'TURNBANK' or swwpmode == 'TURNPHI':

                try:
                    if args[1] == "OFF":
                        acrte.turnbank = -999
                    else:                    
                        # Cap the bank angle between 0 and 90 degrees
                        bankangle = max(0, min(args[1]/ft, 90))
                        acrte.turnbank = bankangle # [deg] desired bank angle
                        
                        # Well, last we defined is turnbank, so append it to the list
                        acrte.last_2_defined.append('turnbank')
                        # Now, unless the first value is also acrte.turnbank, we set that var to -999 again
                        if len(acrte.last_2_defined) > 2:
                            # Reset first variable if one of the remaining 3
                            if acrte.last_2_defined[0] == 'turnhdgr':
                                acrte.turnhdgr = -999.
                            elif acrte.last_2_defined[0] == 'turnspd':
                                acrte.turnspd = -999.
                            elif acrte.last_2_defined[0] == 'turnrad':
                                acrte.turnrad = -999.
                            # Pop the first one
                            acrte.last_2_defined.pop(0)
                        # Lastly, there is a possibility that we are left with a list of duplicates. Fix it
                        if len(acrte.last_2_defined) == 2 and acrte.last_2_defined[0] == acrte.last_2_defined[1]:
                            # Duplicates, pop the first one
                            acrte.last_2_defined.pop(0)
                            
                except:
                    return False, "Error in processing value of turn heading rate"

                # Switch flyturn automatically when this is set
                acrte.swflyby = False
                acrte.swflyturn = True
                
                return True
                    


        # Convert to positions
        name = args[0].upper().strip()

        # Choose reference position ot look up VOR and waypoints
        # First waypoint: own position
        if acrte.nwp == 0:
            reflat = bs.traf.lat[acidx]
            reflon = bs.traf.lon[acidx]

        # Or last waypoint before destination
        else:
            if acrte.wptype[-1] != Route.dest or acrte.nwp == 1:
                reflat = acrte.wplat[-1]
                reflon = acrte.wplon[-1]
            else:
                reflat = acrte.wplat[-2]
                reflon = acrte.wplon[-2]

        # Default altitude, speed and afterwp
        alt     = -999.
        spd     = -999.
        afterwp = ""
        beforewp = ""

        # Is it aspecial take-off waypoint?
        takeoffwpt = name.replace('-', '') == "TAKEOFF"

        # Normal waypoint (no take-off waypoint => see else)
        if not takeoffwpt:

            # Get waypoint position
            success, posobj = txt2pos(name, reflat, reflon)
            if success:
                lat      = posobj.lat
                lon      = posobj.lon

                if posobj.type == "nav" or posobj.type == "apt":
                    wptype = Route.wpnav

                elif posobj.type == "rwy":
                    wptype  = Route.runway

                else:  # treat as lat/lon
                    name    = acid
                    wptype  = Route.wplatlon

                if len(args) > 1 and args[1]:
                    alt = args[1]

                if len(args) > 2 and args[2]:
                    spd = args[2]

                if len(args) > 3 and args[3]:
                    afterwp = args[3]

                if len(args) > 4 and args[4]:
                    beforewp = args[4]

            else:
                return False, "Waypoint " + name + " not found."

        # Take off waypoint: positioned 20% of the runway length after the runway
        else:

            # Look up runway in route
            rwyrteidx = -1
            i      = 0
            while i<acrte.nwp and rwyrteidx<0:
                if acrte.wpname[i].count("/") >0:
                    rwyrteidx = i
                i += 1

            # Only TAKEOFF is specified wihtou a waypoint/runway
            if len(args) == 1 or not args[1]:
                # No runway given: use first in route or current position

                # print ("rwyrteidx =",rwyrteidx)
                # We find a runway in the route, so use it
                if rwyrteidx>0:
                    rwylat   = acrte.wplat[rwyrteidx]
                    rwylon   = acrte.wplon[rwyrteidx]
                    aptidx  = bs.navdb.getapinear(rwylat,rwylon)
                    aptname = bs.navdb.aptname[aptidx]

                    rwyname = acrte.wpname[rwyrteidx].split("/")[1]
                    rwyid = rwyname.replace("RWY","").replace("RW","")
                    rwyhdg = bs.navdb.rwythresholds[aptname][rwyid][2]

                else:
                    rwylat  = bs.traf.lat[acidx]
                    rwylon  = bs.traf.lon[acidx]
                    rwyhdg = bs.traf.trk[acidx]

            elif args[1].count("/") > 0 or len(args) > 2 and args[2]: # we need apt,rwy
                # Take care of both EHAM/RW06 as well as EHAM,RWY18L (so /&, and RW/RWY)
                if args[1].count("/")>0:
                    aptid,rwyname = args[1].split("/")
                else:
                # Runway specified
                    aptid = args[1]
                    rwyname = args[2]

                rwyid = rwyname.replace("RWY", "").replace("RW", "")  # take away RW or RWY
                #                    print ("apt,rwy=",aptid,rwyid)
                # TODO: Add finding the runway heading with rwyrteidx>0 and navdb!!!
                # Try to get it from the database
                try:
                    rwyhdg = bs.navdb.rwythresholds[aptid][rwyid][2]
                except:
                    rwydir = rwyid.replace("L","").replace("R","").replace("C","")
                    try:
                        rwyhdg = float(rwydir)*10.
                    except:
                        return False,name+" not found."

                success, posobj = txt2pos(aptid+"/RW"+rwyid, reflat, reflon)
                if success:
                    rwylat,rwylon = posobj.lat,posobj.lon
                else:
                    rwylat = bs.traf.lat[acidx]
                    rwylon = bs.traf.lon[acidx]

            else:
                return False,"Use ADDWPT TAKEOFF,AIRPORTID,RWYNAME"

            # Create a waypoint 2 nm away from current point
            rwydist = 2.0 # [nm] use default distance away from threshold
            lat,lon = geo.qdrpos(rwylat, rwylon, rwyhdg, rwydist) #[deg,deg
            wptype  = Route.wplatlon

            # Add after the runwy in the route
            if rwyrteidx > 0:
                afterwp = acrte.wpname[rwyrteidx]

            elif acrte.wptype and acrte.wptype[0] == Route.orig:
                afterwp = acrte.wpname[0]

            else:
                # Assume we're called before other waypoints are added
                afterwp = ""

            name = "T/O-" + acid # Use lat/lon naming convention
        # Add waypoint
        wpidx = acrte.addwpt(acidx, name, wptype, lat, lon, alt, spd, afterwp, beforewp)

        # Recalculate flight plan
        acrte.calcfp()

        # Check for success by checking inserted location in flight plan >= 0
        if wpidx < 0:
            return False, "Waypoint " + name + " not added."

        # check for presence of orig/dest
        norig = int(bs.traf.ap.orig[acidx] != "") # 1 if orig is present in route
        ndest = int(bs.traf.ap.dest[acidx] != "") # 1 if dest is present in route

        # Check whether this is first 'real' waypoint (not orig & dest),
        # And if so, make active
        if acrte.nwp - norig - ndest == 1:  # first waypoint: make active
            acrte.direct(acidx, acrte.wpname[norig])  # 0 if no orig
            #print("direct ",self.wpname[norig])
            bs.traf.swlnav[acidx] = True

        if afterwp and acrte.wpname.count(afterwp) == 0:
            print(afterwp, acrte.wpname)
            return True, "Waypoint " + afterwp + " not found\n" + \
                "waypoint added at end of route"
        else:
            return True

    @stack.command
    def addwaypoints(acidx: 'acid', *args):
        ''' ADDWAYPOINTS: Add multiple waypoints to the route of an aircraft.
        
            Arguments:
            - acid: Aircraft id
            - lat_i: Latitude of the ith waypoint
            - lon_i: Longitude of the ith waypoint
            - alt_i: Altitude constraint at the ith waypoint
            - spd_i: Speed constraint at the ith waypoint
            - type_i: Turn type of the ith waypoint. [TURNSPD/TURNRAD/FLYBY]
            - turnspd_i/turnrad_i: Turn speed / turn radius constraint at the ith waypoint
            '''
        # Args come in this order: lat, lon, alt, spd, TURNSPD/TURNRAD/FLYBY, turnspeed or turnrad value
        # If turn is '0', then ignore turnspeed
        if len(args)%6 !=0:
            bs.scr.echo('You missed a waypoint value, arguement number must be a multiple of 6.')
            return

        # Get and reset current aircraft route
        acid = bs.traf.id[acidx]
        acrte = Route._routes.get(acid)

        args = np.reshape(args, (int(len(args)/5), 5))

        for wpdata in args:
            # Get needed values
            lat = float(wpdata[0]) # deg
            lon = float(wpdata[1]) # deg
            if wpdata[2]:
                alt = txt2alt(wpdata[2]) # comes in feet, convert
            else:
                alt = -999
            if wpdata[3]:
                spd = txt2spd(wpdata[3])
            else:
                spd = -999

            # Do flyby or flyturn or flyover processing
            if wpdata[4] == 'FLYTURN': #
                acrte.swflyby   = False
                acrte.swflyturn = True

            elif wpdata[4] == 'FLYBY':
                acrte.swflyby   = True
                acrte.swflyturn = False

            elif wpdata[4] == 'FLYOVER':
                acrte.swflyby   = False
                acrte.swflyturn = False

            else:
                # Either it's a flyby, or a typo.
                acrte.swflyby   = True
                acrte.swflyturn = False

            wpidx = acrte.addwpt_simple(acidx, acid, Route.wplatlon, lat, lon, alt, spd)
        
        # Direct to first waypoint
        acrte.direct(acidx, acrte.wpname[0])  # 0 if no orig
        #print("direct ",self.wpname[norig])
        bs.traf.swlnav[acidx] = True

        # Calculate flight plan
        acrte.calcfp()

        # Check for success by checking inserted location in flight plan >= 0
        if wpidx < 0:
            return False, "Waypoint " + acid + " not added."
        
        # # Direct aircraft to first waypoint
        # acrte.iactwp = 0
        # acrte.direct(acidx, acrte.wpname[0])
        

    def addwpt_simple(self, iac, name, wptype, lat, lon, alt=-999., spd=-999.):
        """Adds waypoint in the most simple way possible"""
        # For safety
        self.nwp = len(self.wplat)

        name = name.upper().strip()

        wplat = lat
        wplon = lon

        # Check if name already exists, if so add integer 01, 02, 03 etc.
        newname = Route.get_available_name(
            self.wpname, name, 3)

        self.addwpt_data(
            False, self.nwp, newname, wplat, wplon, wptype, alt, spd)

        idx = self.nwp
        self.nwp += 1

        #update qdr and "last waypoint switch" in traffic
        if idx>=0:
            bs.traf.actwp.next_qdr[iac] = self.getnextqdr()
            bs.traf.actwp.swlastwp[iac] = (self.iactwp==self.nwp-1)
            
        # Update autopilot settings
        if 0 <= self.iactwp < self.nwp:
            self.direct(iac, self.wpname[self.iactwp])

        return idx

    @stack.command
    @staticmethod
    def before(acidx : 'acid', beforewp: 'wpinroute', addwpt, waypoint, alt: 'alt' = None, spd: 'spd' = None):
        ''' BEFORE acid, wpinroute ADDWPT acid, (wpname/lat,lon),[alt],[spd]

            Before waypoint, add a waypoint to route of aircraft (FMS).
        '''
        return Route.addwptStack(acidx, waypoint, alt, spd, None, beforewp)

    @stack.command
    @staticmethod
    def after(acidx: 'acid', afterwp: 'wpinroute', addwpt, waypoint, alt:'alt' = None, spd: 'spd' = None):
        ''' AFTER acid, wpinroute ADDWPT (wpname/lat,lon),[alt],[spd]

            After waypoint, add a waypoint to route of aircraft (FMS).
        '''
        return Route.addwptStack(acidx, waypoint, alt, spd, afterwp)

    @stack.command
    @staticmethod
    def at(acidx: 'acid', atwp : 'wpinroute', *args):
        ''' AT acid, wpinroute [DEL] ALT/SPD/DO alt/spd/stack command'''
        acid = bs.traf.id[acidx]
        acrte = Route._routes[acid]
        if atwp in acrte.wpname:
            wpidx = acrte.wpname.index(atwp)

            if not args or \
                    (len(args) == 1 and not args[0].count("/") == 1):
                # Only show Altitude and/or speed set in route at this waypoint:
                #    KL204 AT LOPIK => acid AT wpinroute: show alt & spd constraints at this waypoint
                #    KL204 AT LOPIK SPD => acid AT wpinroute SPD: show spd constraint at this waypoint
                #    KL204 AT LOPIK ALT => acid AT wpinroute ALT: show alt constraint at this waypoint
                txt = atwp + " : "

                # Select what to show
                if len(args) == 0:
                    swalt = True
                    swspd = True
                    swat  = True
                else:
                    swalt = args[0].upper() == "ALT"
                    swspd = args[0].upper() in ("SPD","SPEED")
                    swat  = args[0].upper() in ("DO", "STACK")

                    # To be safe show both when we do not know what
                    if not (swalt or swspd or swat):
                        swalt = True
                        swspd = True
                        swat  = True

                # Show altitude
                if swalt:
                    if acrte.wpalt[wpidx] < 0:
                        txt += "-----"

                    elif acrte.wpalt[wpidx] > 4500 * ft:
                        fl = int(round((acrte.wpalt[wpidx] / (100. * ft))))
                        txt += "FL" + str(fl)

                    else:
                        txt += str(int(round(acrte.wpalt[wpidx] / ft)))

                    if swspd:
                        txt += "/"

                # Show speed
                if swspd:
                    if acrte.wpspd[wpidx] < 0:
                        txt += "---"
                    else:
                        txt += str(int(round(acrte.wpspd[wpidx] / kts)))

                # Type
                if swalt and swspd:
                    if acrte.wptype[wpidx] == Route.orig:
                        txt += "[orig]"
                    elif acrte.wptype[wpidx] == Route.dest:
                        txt += "[dest]"

                # Show also stacked commands for when passing this waypoint
                if swat:
                    if len(acrte.wpstack[wpidx])>0:
                        txt = txt+"\nStack:\n"
                        for stackedtxt in acrte.wpstack[wpidx]:
                            txt = txt + stackedtxt + "\n"


                return True, txt

            elif args[0].count("/")==1:
                # Set both alt & speed at this waypoint
                #     KL204 AT LOPIK FL090/250  => acid AT wpinroute alt/spd
                success = True

                # Use parse from stack.py to interpret alt & speed
                alttxt, spdtxt = args[0].split('/')

                # Edit waypoint altitude constraint
                if alttxt.count('-') > 1: # "----" = delete
                    acrte.wpalt[wpidx]  = -999.
                else:
                    try:
                        acrte.wpalt[wpidx] = txt2alt(alttxt)
                        acrte.calcfp()   # Recalculate VNAV axes
                    except ValueError as e:
                        success = False

                # Edit waypoint speed constraint
                if spdtxt.count('-') > 1: # "----" = delete
                    acrte.wpspd[wpidx]  = -999.
                else:
                    try:
                        acrte.wpalt[wpidx] = txt2spd(spdtxt)
                    except ValueError as e:
                        success = False

                if not success:
                    return False,"Could not parse "+args[0]+" as alt / spd"

                # If success: update flight plan and guidance
                acrte.calcfp()
                acrte.direct(acidx, acrte.wpname[acrte.iactwp])

            #acid AT wpinroute ALT/SPD alt/spd
            elif len(args)>=2:
                # KL204 AT LOPIK ALT FL090 => set altitude to be reached at this waypoint in route
                # KL204 AT LOPIK SPD 250 => Set speed at twhich is set at this waypoint
                # KL204 AT LOPIK DO PAN LOPIK => When passing stack command after DO
                # KL204 AT LOPIK STACK PAN LOPIK => AT...STACK synonym for AT...DO
                # KL204 AT LOPIK DO ALT FL240 => => stack "KL204 ALT FL240" => use acid from beginning if omitted as first argument

                swalt = args[0].upper()=="ALT"
                swspd = args[0].upper() in ("SPD","SPEED")
                swat  = args[0].upper() in ("DO","STACK")

                # Use parse from stack.py to interpret alt & speed

                # Edit waypoint altitude constraint
                if swalt:
                    try:
                        acrte.wpalt[wpidx] = txt2alt(args[1])
                    except ValueError as e:
                        return False, e.args[0]

                # Edit waypoint speed constraint
                elif swspd:
                    try:
                        acrte.wpspd[wpidx] = txt2spd(args[1])
                    except ValueError as e:
                        return False, e.args[0]

                # add stack command: args[1] is DO or STACK, args[2:] contains a command
                elif swat:
                    # Check if first argument is missing aircraft id, if so, use this acid

                    # IF command starts with aircraft id, it is not missing
                    cmd = args[1].upper()
                    if not(cmd in bs.traf.id):
                        # Look up arg types
                        try:
                            cmdobj = Command.cmddict[cmd]

                            # Command found, check arguments
                            argtypes = cmdobj.annotations

                            if len(argtypes)>0 and argtypes[0]=="acid" and not (len(args)>2 and args[2].upper() in bs.traf.id):
                                # missing acid, so add ownship acid
                                acrte.wpstack[wpidx].append(acid+" "+" ".join(args[1:]))
                            else:
                                # This command does not need an acid or it is already first argument
                                acrte.wpstack[wpidx].append(" ".join(args[1:]))
                        except:
                            return False, "Stacked command "+cmd+" unknown or syntax error"
                    else:
                        # Command line starts with an aircraft id at the beginning of the command line, stack it
                        acrte.wpstack[wpidx].append(" ".join(args[1:]))

                # Delete a constraint (or both) at this waypoint
                elif args[0]=="DEL" or args[0]=="DELETE" or args[0]=="CLR" or args[0]=="CLEAR" :
                    swalt = args[1].upper()=="ALT"
                    swspd = args[1].upper() in ("SPD","SPEED")
                    swboth  = args[1].upper()=="BOTH"
                    swall   = args[1].upper()=="ALL"

                    if swspd or swboth or swall:
                        acrte.wpspd[wpidx]  = -999.

                    if swalt or swboth or swall:
                        acrte.wpalt[wpidx]  = -999.

                    if swall:
                        acrte.wpstack[wpidx]=[]

                else:
                    return False,"No "+args[0]+" at ",atwp


                # If success: update flight plan and guidance
                acrte.calcfp()
                acrte.direct(acidx, acrte.wpname[acrte.iactwp])

        # Waypoint not found in route
        else:
            return False, atwp + " not found in route " + acid

        return True

    def overwrite_wpt_data(self, wpidx, wpname, wplat, wplon, wptype, wpalt,
                           wpspd):
        """
        Overwrites information for a waypoint, via addwpt_data/9
        """

        self.addwpt_data(True, wpidx, wpname, wplat, wplon, wptype, wpalt,
                         wpspd)

    def insert_wpt_data(self, wpidx, wpname, wplat, wplon, wptype, wpalt,
                        wpspd):
        """
        Inserts information for a waypoint, via addwpt_data/9
        """

        self.addwpt_data(False, wpidx, wpname, wplat, wplon, wptype, wpalt,
                         wpspd)

    def addwpt_data(self, overwrt, wpidx, wpname, wplat, wplon, wptype,
                    wpalt, wpspd):
        """
        Overwrites or inserts information for a waypoint
        """
        wplat = (wplat + 90.) % 180. - 90.
        wplon = (wplon + 180.) % 360. - 180.

        if overwrt:
            self.wpname[wpidx]  = wpname
            self.wplat[wpidx]   = wplat
            self.wplon[wpidx]   = wplon
            self.wpalt[wpidx]   = wpalt
            self.wpspd[wpidx]   = wpspd
            self.wptype[wpidx]  = wptype
            self.wpflyby[wpidx] = self.swflyby
            self.wpflyturn[wpidx] = self.swflyturn
            self.wpturnbank[wpidx] = self.turnbank
            self.wpturnrad[wpidx] = self.turnrad
            self.wpturnspd[wpidx] = self.turnspd
            self.wpturnhdgr[wpidx] = self.turnhdgr
            self.wprta[wpidx]   = -999.0 # initially no RTA
            self.wpstack[wpidx] = []

        else:
            self.wpname.insert(wpidx, wpname)
            self.wplat.insert(wpidx, wplat)
            self.wplon.insert(wpidx, wplon)
            self.wpalt.insert(wpidx, wpalt)
            self.wpspd.insert(wpidx, wpspd)
            self.wptype.insert(wpidx, wptype)
            self.wpflyby.insert(wpidx, self.swflyby)
            self.wpflyturn.insert(wpidx, self.swflyturn)
            self.wpturnbank.insert(wpidx,self.turnbank)
            self.wpturnrad.insert(wpidx, self.turnrad)
            self.wpturnspd.insert(wpidx, self.turnspd)
            self.wpturnhdgr.insert(wpidx, self.turnhdgr)
            self.wprta.insert(wpidx,-999.0)       # initially no RTA
            self.wpstack.insert(wpidx,[])


    def addwpt(self, iac, name, wptype, lat, lon, alt=-999., spd=-999., afterwp="", beforewp=""):
        """Adds waypoint an returns index of waypoint, lat/lon [deg], alt[m]"""
#        print ("addwpt:")
#        print ("iac = ",iac)
#        print ("name = "+name)
#        print ("alt = ",alt)
#        print ("spd = ",spd)
#        print ("afterwp ="+afterwp)
#        print
        # For safety
        self.nwp = len(self.wplat)

        name = name.upper().strip()

        wplat = lat
        wplon = lon

        # Be default we trust, distrust needs to be earned
        wpok = True   # switch for waypoint check

        # Check if name already exists, if so add integer 01, 02, 03 etc.
        wprtename = Route.get_available_name(
            self.wpname, name)
        # Select on wptype
        # ORIGIN: Wptype is origin/destination?
        if wptype == Route.orig or wptype == Route.dest:
            orig = wptype == Route.orig
            wpidx = 0 if orig else -1
            suffix = "ORIG" if orig else "DEST"

            if not name == bs.traf.id[iac] + suffix:  # published identifier
                i = bs.navdb.getaptidx(name)
                if i >= 0:
                    wplat = bs.navdb.aptlat[i]
                    wplon = bs.navdb.aptlon[i]

            if not orig and alt < 0:
                alt = 0

            # Overwrite existing origin/dest
            if self.nwp > 0 and self.wptype[wpidx] == wptype:
                self.overwrite_wpt_data(
                    wpidx, wprtename, wplat, wplon, wptype, alt, spd)

            # Or add before first waypoint/append to end
            else:
                if not orig:
                    wpidx = len(self.wplat)

                self.insert_wpt_data(
                    wpidx, wprtename, wplat, wplon, wptype, alt, spd)

                self.nwp += 1
                if orig and self.iactwp >= 0:
                    self.iactwp += 1
                elif not orig and self.iactwp < 0 and self.nwp == 1:
                    # When only waypoint: adjust pointer to point to destination
                    self.iactwp = 0

            idx = 0 if orig else self.nwp - 1

        # NORMAL: Wptype is normal waypoint? (lat/lon or nav)
        else:
            # Lat/lon: wpname is then call sign of aircraft: add number
            if wptype == Route.wplatlon:
                newname = Route.get_available_name(
                    self.wpname, name, 3)

            # Else make data complete with nav database and closest to given lat,lon
            else: # so wptypewpnav
                newname = wprtename

                if not wptype == Route.runway:
                    i = bs.navdb.getwpidx(name, lat, lon)
                    wpok = (i >= 0)

                    if wpok:
                        wplat = bs.navdb.wplat[i]
                        wplon = bs.navdb.wplon[i]
                    else:
                        i = bs.navdb.getaptidx(name)
                        wpok = (i >= 0)
                        if wpok:
                            wplat = bs.navdb.aptlat[i]
                            wplon = bs.navdb.aptlon[i]

            # Check if afterwp or beforewp is specified and found:
            aftwp = afterwp.upper().strip()  # Remove space, upper case
            bfwp = beforewp.upper().strip()

            if wpok:

                if (afterwp and self.wpname.count(aftwp) > 0) or \
                        (beforewp and self.wpname.count(bfwp) > 0):

                    wpidx = self.wpname.index(aftwp) + 1 if afterwp else \
                        self.wpname.index(bfwp)

                    self.insert_wpt_data(
                        wpidx, newname, wplat, wplon, wptype, alt, spd)

                    if afterwp and self.iactwp >= wpidx:
                        self.iactwp += 1

                # No afterwp: append, just before dest if there is a dest
                else:
                    # Is there a destination?
                    if self.nwp > 0 and self.wptype[-1] == Route.dest:
                        wpidx = self.nwp - 1
                    else:
                        wpidx = self.nwp

                    self.addwpt_data(
                        False, wpidx, newname, wplat, wplon, wptype, alt, spd)

                idx = wpidx
                self.nwp += 1

            else:
                idx = -1
                if len(self.wplat) == 1:
                    self.iactwp = 0

        #update qdr and "last waypoint switch" in traffic
        if idx>=0:
            bs.traf.actwp.next_qdr[iac] = self.getnextqdr()
            bs.traf.actwp.swlastwp[iac] = (self.iactwp==self.nwp-1)

        # Update waypoints
        if not (wptype == Route.calcwp):
            self.calcfp()

        # Update autopilot settings
        # If we added a waypoint but iactwp is still -1, make it 0
        if wpok and self.iactwp < 0:
            self.iactwp = 0
            
        if wpok and 0 <= self.iactwp < self.nwp:
            self.direct(iac, self.wpname[self.iactwp])

        return idx

    @stack.command(aliases=("DIRECTTO", "DIRTO"))
    @staticmethod
    def direct(acidx: 'acid', wpname: 'wpinroute'):
        """DIRECT acid wpname
        
            Go direct to specified waypoint in route (FMS)"""
        acid = bs.traf.id[acidx]
        acrte = Route._routes[acid]
        wpidx = acrte.wpname.index(wpname)

        acrte.iactwp = wpidx
        bs.traf.actwp.lat[acidx]    = acrte.wplat[wpidx]
        bs.traf.actwp.lon[acidx]    = acrte.wplon[wpidx]
        bs.traf.actwp.flyby[acidx]  = acrte.wpflyby[wpidx]
        bs.traf.actwp.flyturn[acidx] = acrte.wpflyturn[wpidx]
        bs.traf.actwp.turnbank[acidx] = acrte.wpturnbank[wpidx]
        bs.traf.actwp.turnrad[acidx] = acrte.wpturnrad[wpidx]
        bs.traf.actwp.turnspd[acidx] = acrte.wpturnspd[wpidx]
        bs.traf.actwp.turnhdgr[acidx] = acrte.wpturnhdgr[wpidx]

        bs.traf.actwp.nextturnlat[acidx], bs.traf.actwp.nextturnlon[acidx], \
        bs.traf.actwp.nextturnspd[acidx], bs.traf.actwp.nextturnrad[acidx], \
        bs.traf.actwp.nextturnhdgr[acidx] , bs.traf.actwp.nextturnidx[acidx]\
            = acrte.getnextturnwp()

        # Determine next turn waypoint data

        # Do calculation for VNAV
        acrte.calcfp()

        bs.traf.actwp.xtoalt[acidx] = acrte.wpxtoalt[wpidx]
        bs.traf.actwp.nextaltco[acidx] = acrte.wptoalt[wpidx]

        bs.traf.actwp.torta[acidx]    = acrte.wptorta[wpidx]    # available for active RTA-guidance
        bs.traf.actwp.xtorta[acidx]  = acrte.wpxtorta[wpidx]  # available for active RTA-guidance

        #VNAV calculations like V/S and speed for RTA
        bs.traf.ap.ComputeVNAV(acidx, acrte.wptoalt[wpidx], acrte.wpxtoalt[wpidx],\
                                    acrte.wptorta[wpidx],acrte.wpxtorta[wpidx])

        # If there is a speed specified, process it
        if acrte.wpspd[wpidx]>0.:
            # Set target speed for autopilot

            if acrte.wpalt[wpidx] < 0.0:
                alt = bs.traf.alt[acidx]
            else:
                alt = acrte.wpalt[wpidx]

            # Check for valid Mach or CAS
            if acrte.wpspd[wpidx] <2.0:
                cas = mach2cas(acrte.wpspd[wpidx], alt)
            else:
                cas = acrte.wpspd[wpidx]

            # Save it for next leg
            bs.traf.actwp.nextspd[acidx] = cas

        # No speed specified for next leg
        else:
            bs.traf.actwp.nextspd[acidx] = -999.


        qdr_,dist_ = geo.qdrdist(bs.traf.lat[acidx], bs.traf.lon[acidx],
                             bs.traf.actwp.lat[acidx], bs.traf.actwp.lon[acidx])

        # Save leg length & direction in actwp data
        bs.traf.actwp.curlegdir[acidx] = qdr_      #[deg]
        bs.traf.actwp.curleglen[acidx] = dist_*nm  #[m]
        
        if acrte.wpflyturn[wpidx] and acrte.wpturnrad[wpidx]>0.: # turn radius specified
            turnrad = acrte.wpturnrad[wpidx]
        # Overwrite is hdgrate  defined
        if acrte.wpflyturn[wpidx] and acrte.wpturnhdgr[wpidx] > 0.: # heading rate specified
            turnrad = bs.traf.tas[acidx]*360./(2*np.pi*acrte.wpturnhdgr[wpidx])
        
        # Update turndist so ComputeVNAV works, is there a next leg direction or not?
        if bs.traf.actwp.next_qdr[acidx] < -900.:
            local_next_qdr = qdr_
        else:
            local_next_qdr = bs.traf.actwp.next_qdr[acidx]

        # Calculate turn dist (and radius which we do not use now, but later) now for scalar variable [acidx]
        bs.traf.actwp.turndist[acidx], turnrad, turnspd, turnbank, turnhdgr = \
            bs.traf.actwp.calcturn(acidx, bs.traf.tas[acidx], qdr_, 
                                    local_next_qdr, bs.traf.actwp.turnbank[acidx], 
                                    bs.traf.actwp.turnrad[acidx],bs.traf.actwp.turnspd[acidx] ,
                                    bs.traf.actwp.turnhdgr[acidx], bs.traf.actwp.flyturn[acidx],
                                    bs.traf.actwp.flyby[acidx])  # update turn distance for VNAV

        # Update turn data
        bs.traf.actwp.turnrad[acidx]      = turnrad
        bs.traf.actwp.turnspd[acidx]      = turnspd
        bs.traf.actwp.turnhdgr[acidx]     = turnhdgr
        bs.traf.ap.turnphi[acidx] = np.deg2rad(turnbank)

        bs.traf.swlnav[acidx] = True
        return True

    @stack.command(name='RTA')
    @staticmethod
    def SetRTA(acidx: 'acid', wpname: 'wpinroute', time: 'time'):  # all arguments of setRTA
        """ RTA acid, wpname, time
        
            Add RTA to waypoint record"""
        acid = bs.traf.id[acidx]
        acrte = Route._routes[acid]
        wpidx = acrte.wpname.index(wpname)
        acrte.wprta[wpidx] = time

        # Recompute route and update actwp because of RTA addition
        acrte.direct(acidx, acrte.wpname[acrte.iactwp])

        return True

    @stack.command
    @staticmethod
    def listrte(acidx: 'acid', ipagetxt="0"):
        """ LISTRTE acid, [pagenr]

            Show list of route in window per page of 5 waypoints/"""
        # First get the appropriate ac route
        ipage = int(ipagetxt)
        acid = bs.traf.id[acidx]
        acrte = Route._routes[acid]
        if acrte.nwp <= 0:
            return False, "Aircraft has no route."

        for i in range(ipage * 7, ipage * 7 + 7):
            if 0 <= i < acrte.nwp:
                # Name
                if i == acrte.iactwp:
                    txt = "*" + acrte.wpname[i] + " : "
                else:
                    txt = " " + acrte.wpname[i] + " : "

                # Altitude
                if acrte.wpalt[i] < 0:
                    txt += "-----/"

                elif acrte.wpalt[i] > 4500 * ft:
                    fl = int(round((acrte.wpalt[i] / (100. * ft))))
                    txt += "FL" + str(fl) + "/"

                else:
                    txt += str(int(round(acrte.wpalt[i] / ft))) + "/"

                # Speed
                if acrte.wpspd[i] < 0.:
                    txt += "---"
                elif acrte.wpspd[i] > 2.0:
                    txt += str(int(round(acrte.wpspd[i] / kts)))
                else:
                    txt += "M" + str(acrte.wpspd[i])

                # Type: orig, dest, C = flyby, | = flyover, U = flyturn
                if acrte.wptype[i] == Route.orig:
                    txt += "[orig]"
                elif acrte.wptype[i] == Route.dest:
                    txt += "[dest]"
                elif acrte.wpflyturn[i]:
                    txt += "[U]"
                elif acrte.wpflyby[i]:
                    txt += "[C]"
                else: # FLYOVER
                    txt += "[|]"


                # Display message
                bs.scr.echo(txt)

        # Add command for next page to screen command line
        npages = int((acrte.nwp + 6) / 7)
        if ipage + 1 < npages:
            bs.scr.cmdline("LISTRTE " + acid + "," + str(ipage + 1))

    def getnextturnwp(self):
        """Give the next turn waypoint data."""
        # Starting point
        wpidx = self.iactwp
        # Find next turn waypoint index
        turnidx_all = np.where(self.wpflyturn)[0]
        argwhere_arr = np.argwhere(turnidx_all>=wpidx)
        if argwhere_arr.size == 0:
            # No turn waypoints, return default values
            return [0., 0., -999., -999., -999, -999.]

        trnidx = turnidx_all[np.argwhere(turnidx_all>=wpidx)[0]][0]

        i = bs.traf.id.index(self.acid)
        # Calculate the turn first. We need to assume  that the aircraft is
        # coming from perfectly on the previous leg.
        qdr, _ = geo.qdrdist(self.wplat[trnidx-1], self.wplon[trnidx-1],
                                self.wplat[trnidx], self.wplon[trnidx])
        if trnidx < len(self.wplat)-1:
            local_next_qdr, _ = geo.qdrdist(self.wplat[trnidx], self.wplon[trnidx],
                                self.wplat[trnidx+1], self.wplon[trnidx+1])
        else:
            local_next_qdr = qdr
            
        turndist, turnrad, turnspd, turnbank, turnhdgr = \
        bs.traf.actwp.calcturn(i, bs.traf.tas[i], qdr, 
                    local_next_qdr, self.wpturnbank[trnidx], 
                    self.wpturnrad[trnidx],self.wpturnspd[trnidx],self.wpturnhdgr[trnidx], True, False)
        
        return [self.wplat[trnidx], self.wplon[trnidx], turnspd, turnrad, turnhdgr, trnidx]

    def getnextwp(self):
        """Go to next waypoint and return data"""

        if self.flag_landed_runway:

            # when landing, LNAV is switched off
            lnavon = False

            # no further waypoint
            nextqdr = -999.

            # and the aircraft just needs a fixed heading to
            # remain on the runway
            # syntax: HDG acid,hdg (deg,True)
            name = self.wpname[self.iactwp]

            # Change RW06,RWY18C,RWY24001 to resp. 06,18C,24
            if "RWY" in name:
                rwykey = name[8:10]
                if len(name)>10:
                    if not name[10].isdigit():
                        rwykey = name[8:11]
            # also if it is only RW
            else:
                rwykey = name[7:9]
                if len(name) > 9:
                    if not name[9].isdigit():
                        rwykey = name[7:10]

            # Use this code to look up runway heading
            wphdg = bs.navdb.rwythresholds[name[:4]][rwykey][2]

            # keep constant runway heading
            stack.stack("HDG " + str(self.acid) + " " + str(wphdg))

            # start decelerating
            stack.stack("DELAY " + "10 " + "SPD " + str(self.acid) + " " + "10")

            # delete aircraft
            stack.stack("DELAY " + "42 " + "DEL " + str(self.acid))

            swlastwp = (self.iactwp == self.nwp - 1)

            return self.wplat[self.iactwp],self.wplon[self.iactwp],   \
                           self.wpalt[self.iactwp],self.wpspd[self.iactwp],   \
                           self.wpxtoalt[self.iactwp],self.wptoalt[self.iactwp], \
                           self.wpxtorta[self.iactwp], self.wptorta[self.iactwp], \
                           lnavon,self.wpflyby[self.iactwp], \
                           self.wpflyturn[self.iactwp],self.wpturnrad[self.iactwp], \
                           self.wpturnspd[self.iactwp], self.wpturnhdgr[self.iactwp], \
                           self.wpturnbank[self.iactwp], nextqdr, swlastwp

        # Switch LNAV off when last waypoint has been passed
        lnavon = self.iactwp < self.nwp -1

        # if LNAV on: increase counter
        if lnavon:
            self.iactwp += 1

        # Activate switch to indicate that this is the last waypoint (for lenient passing logic in actwp.Reached function)
        swlastwp = (self.iactwp == self.nwp-1)

        nextqdr = self.getnextqdr()

        # in case that there is a runway, the aircraft should remain on it
        # instead of deviating to the airport centre
        # When there is a destination: current = runway, next  = Dest
        # Else: current = runway and this is also the last waypoint
        if (self.wptype[self.iactwp] == 5 and
                self.wpname[self.iactwp] == self.wpname[-1]) or \
           (self.wptype[self.iactwp] == 5 and self.iactwp+1<self.nwp and
                self.wptype[self.iactwp + 1] == 3):

            self.flag_landed_runway = True

        #print ("getnextwp:",self.wpname[self.iactwp],"   torta = ",self.wptorta[self.iactwp])


        return self.wplat[self.iactwp],self.wplon[self.iactwp],   \
                self.wpalt[self.iactwp],self.wpspd[self.iactwp],   \
                self.wpxtoalt[self.iactwp],self.wptoalt[self.iactwp], \
                self.wpxtorta[self.iactwp], self.wptorta[self.iactwp], \
                lnavon,self.wpflyby[self.iactwp], \
                self.wpflyturn[self.iactwp],self.wpturnrad[self.iactwp], \
                self.wpturnspd[self.iactwp], self.wpturnhdgr[self.iactwp], \
                self.wpturnbank[self.iactwp], nextqdr, swlastwp

    def runactwpstack(self):
        for cmdline in self.wpstack[self.iactwp]:
            stack.stack(cmdline)
            #debug
            # stack.stack("ECHO "+self.acid+" AT "+self.wpname[self.iactwp]+" command issued:"+cmdline)
        return

    @stack.command(aliases=("DELROUTE",))
    @staticmethod
    def delrte(acidx: 'acid' = None):
        """ DELRTE acid
            Delete for this a/c the complete route/dest/orig (FMS)."""
        if acidx is None:
            if bs.traf.ntraf == 0:
                return False, 'No aircraft in simulation'
            if bs.traf.ntraf > 1:
                return False, 'Specify callsign of aircraft to delete route of'
            acidx = 0
        # Simple re-initialize this route as empty
        acid = bs.traf.id[acidx]
        acrte = Route._routes[acid]
        acrte.__init__(acid)

        # Also disable LNAV,VNAV if route is deleted
        bs.traf.swlnav[acidx]    = False
        bs.traf.swvnav[acidx]    = False
        bs.traf.swvnavspd[acidx] = False

        return True

    @stack.command(aliases=("DELWP",))
    @staticmethod
    def delwpt(acidx: 'acid', wpname: 'wpinroute'):
        """DELWPT acid,wpname
        
           Delete a waypoint from a route (FMS). """
        # Delete complete route?
        if wpname == "*":
            return Route.delrte(acidx)

        # Look up waypoint
        acid = bs.traf.id[acidx]
        acrte = Route._routes[acid]
        try:
            wpidx = acrte.wpname.index(wpname.upper())
        except ValueError:
            return False, "Waypoint " + wpname + " not found"
        # check if active way point is the one being deleted and that it is not the last wpt.
        # If active wpt is deleted then change path of aircraft
        if acrte.iactwp == wpidx and not wpidx == acrte.nwp - 1:
            acrte.direct(acidx, acrte.wpname[wpidx + 1])

        acrte.nwp =acrte.nwp - 1
        del acrte.wpname[wpidx]
        del acrte.wplat[wpidx]
        del acrte.wplon[wpidx]
        del acrte.wpalt[wpidx]
        del acrte.wpspd[wpidx]
        del acrte.wprta[wpidx]
        del acrte.wptype[wpidx]
        if acrte.iactwp > wpidx:
            acrte.iactwp = max(0, acrte.iactwp - 1)

        acrte.iactwp = min(acrte.iactwp, acrte.nwp - 1)

        # If no waypoints left, make sure to disable LNAV/VNAV
        if acrte.nwp==0 and (acidx or acidx==0):
            bs.traf.swlnav[acidx]    =  False
            bs.traf.swvnav[acidx]    =  False
            bs.traf.swvnavspd[acidx] =  False

        return True


    def insertcalcwp(self, i, name):
        """Insert empty wp with no attributes at location i"""

        self.wpname.insert(i,name)
        self.wplat.insert(i,0.)
        self.wplon.insert(i,0.)
        self.wpalt.insert(i,-999.)
        self.wpspd.insert(i,-999.)
        self.wptype.insert(i,Route.calcwp)

    def calcfp(self): # Current Flight Plan calculations, which actualize based on flight condition
        """Do flight plan calculations"""

        # Note: No Top of Descent or Top of Climb can inserted here
        # as this depends on the speed, which might be undefined (often is)
        # Guidance in autpilot.py takes care of ToD and ToC logic while flying using current speed
        # This routine prepares data for this by adding a "ruler" along the flight plan in the form of
        # distance at wp to next altitude constraint (xtoalt), its index ial and the value (toalt)
        # same logic is used for time consarint (requieed time of arrival) RTAs at waypoints

        # Direction to waypoint
        self.nwp = len(self.wpname)

        # Create cleared flight plan calculation table
        self.wpdirfrom   = self.nwp*[0.]  # [deg] Direction of leg laving this waypoint
        self.wpdirto     = self.nwp*[0.]  # [deg] Direction of leg ot this waypoint (if it exists)
        self.wpdistto    = self.nwp*[0.]  # [nm] Distance of leg to this waypoint in nm
        self.wpialt      = self.nwp*[-1]  # wp index of next alttud constraint
        self.wptoalt     = self.nwp*[-999.] # [m] next alt contraint
        self.wpxtoalt    = self.nwp*[1.]  # [m] dist to next alt constraint, default 1.0 to avoid division by zero
        self.wpirta      = self.nwp*[-1]  # wp index of next time constraint
        self.wptorta     = self.nwp*[-999.] # [s] next time constraint
        self.wpxtorta    = self.nwp*[1.]  # [m] dist to next time constraint, default 1.0 to avoid division by zero

        # No waypoints: make empty variables to be safe and return: nothing to do
        if self.nwp==0:
            return

        # Calculate lateral leg data
        # LNAV: Calculate leg distances and directions

        for i in range(0, self.nwp - 1):
            qdr,dist = geo.qdrdist(self.wplat[i]  ,self.wplon[i],
                                self.wplat[i+1],self.wplon[i+1])
            self.wpdirfrom[i] = qdr    # [deg]
            self.wpdistto[i+1]  = dist #[nm]  distto is in nautical miles

        # Also add "from direction" as to directions so no need to shift for actwpdata
        # direction to will be overwritten in actwpdata in case of a direct to
        # Add current pos to first waypoint as default value for direction to 1st waypoint
        iac = bs.traf.id2idx(self.acid)
        qdr,dist = geo.qdrdist(bs.traf.lat[iac],bs.traf.lon[iac],
                               self.wplat[0],self.wplon[0])
        self.wpdirto = [qdr]+self.wpdirfrom[0:-1] #[deg] Direction to waypoints

        # Continue flying in the saem direction
        if self.nwp>1:
            self.wpdirfrom[-1] = self.wpdirfrom[-2]

        # Calculate longitudinal leg data
        # VNAV: calc next altitude constraint: index, altitude and distance to it
        ialt = -1     # index to waypoint with next altitude constraint
        toalt = -999. # value of next altitude constraint
        xtoalt = 0.   # distance to next altitude constraint from this wp
        for i in range(self.nwp-1,-1,-1):

            # waypoint with altitude constraint (dest of al specified)
            if self.wptype[i]==Route.dest:
                ialt   = i
                toalt  = 0.
                xtoalt = 0.                # [m]

            elif self.wpalt[i] >= 0:
                ialt   = i
                toalt  = self.wpalt[i]
                xtoalt = 0.                # [m]

            # waypoint with no altitude constraint:keep counting
            else:
                if i!=self.nwp-1:
                    xtoalt = xtoalt + self.wpdistto[i+1]*nm  # [m] xtoalt is in meters!
                else:
                    xtoalt = 0.0

            self.wpialt[i]   = ialt
            self.wptoalt[i]  = toalt   #[m]
            self.wpxtoalt[i] = xtoalt  #[m]

        # RTA: calc next rta constraint: index, altitude and distance to it
        # If any RTA.
        if any(np.array(self.wprta)>=0.0):
            #print("Yes, I found RTAs")
            irta = -1       # index of wp
            torta = -999.   # next rta value
            xtorta = 0.     # distance to next rta
            for i in range(self.nwp - 1, -1, -1):

                # waypoint with rta: reset counter, update rts
                if self.wprta[i] >= 0:
                    irta = i
                    torta = self.wprta[i]
                    xtorta = 0.  # [m]

                # waypoint with no altitude constraint:keep counting
                else:
                    if i != self.nwp - 1:
                        # No speed or rta constraint: add to xtorta
                        if self.wpspd[i] <= 0.0:
                            xtorta = xtorta + self.wpdistto[i + 1] * nm  # [m] xtoalt is in meters!
                        else:
                            # speed constraint on this leg: shift torta to account for this
                            # altitude unknown
                            if self.wptoalt[i] >0.:
                                alt = toalt
                            else:
                                # TODO: current a/c altitude would be better guess, but not accessible here
                                # as we do not know aircraft index for this route
                                alt = 10000.*ft # default to minimize errors, when no alt constraints are present
                            legtas = casormach2tas(self.wpspd[i],alt)
                            #TODO: account for wind at this position vy adding wind vectors to waypoints?

                            # xtorta stays the same! This leg will not be available for RTA scheduling, so distance
                            # is not in xtorta. Therefore we need to subtract legtime to ignore this leg for the RTA
                            # scheduling
                            legtime = self.wpdistto[i+1]/legtas
                            torta = torta - legtime
                    else:
                        xtorta = 0.0
                        torta = -999.0

                self.wpirta[i]   = irta
                self.wptorta[i]  = torta  # [s]
                self.wpxtorta[i] = xtorta  # [m]
            #print("wpxtorta=",self.wpxtorta)
            #print("wptorta=", self.wptorta)

    def findact(self,i):
        """ Find best default active waypoint.
        This function is called during route creation"""
        #print "findact is called.!"

        # Check for easy answers first
        if self.nwp<=0:
            return -1

        elif self.nwp == 1:
            return 0

        # Find closest
        wplat  = np.array(self.wplat)
        wplon  = np.array(self.wplon)
        dy = (wplat - bs.traf.lat[i])
        dx = (wplon - bs.traf.lon[i]) * bs.traf.coslat[i]
        dist2 = dx*dx + dy*dy
        # Note: the max() prevents walking back, even in cases when this might be apropriate,
        # such as when previous waypoints have been deleted

        iwpnear = max(self.iactwp, np.argmin(dist2))

        #Unless behind us, next waypoint?
        if iwpnear+1<self.nwp:
            qdr = math.degrees(math.atan2(dx[iwpnear],dy[iwpnear]))
            delhdg = abs(degto180(bs.traf.trk[i]-qdr))

            # we only turn to the first waypoint if we can reach the required
            # heading before reaching the waypoint
            time_turn = max(0.01,bs.traf.tas[i])*math.radians(delhdg)/(g0*math.tan(bs.traf.ap.bankdef[i]))
            time_straight= math.sqrt(dist2[iwpnear])*60.*nm/max(0.01,bs.traf.tas[i])

            if time_turn > time_straight:
                iwpnear += 1

        return iwpnear

    @stack.command
    @staticmethod
    def dumprte(acidx: 'acid'):
        """ DUMPRTE acid

            Write route to output/routelog.txt.
        """
        acid = bs.traf.id[acidx]
        acrte = Route._routes[acid]
        # Open file in append mode, write header
        with open(bs.resource(bs.settings.log_path) / 'routelog.txt', "a") as f:
            f.write("\nRoute "+acid+":\n")
            f.write("(name,type,lat,lon,alt,spd,toalt,xtoalt)  ")
            f.write("type: 0=latlon 1=navdb  2=orig  3=dest  4=calwp\n")

            # write flight plan VNAV data (Lateral is visible on screen)
            for j in range(acrte.nwp):
                f.write( str(( j, acrte.wpname[j], acrte.wptype[j],
                      round(acrte.wplat[j], 4), round(acrte.wplon[j], 4),
                      int(0.5+acrte.wpalt[j]/ft), int(0.5+acrte.wpspd[j]/kts),
                      int(0.5+acrte.wptoalt[j]/ft), round(acrte.wpxtoalt[j]/nm, 3)
                      )) + "\n")

            # End of data
            f.write("----\n")
            f.close()

    def getnextqdr(self):
        # get qdr for next leg
        if -1 < self.iactwp < self.nwp - 1:
            nextqdr, dist = geo.qdrdist(\
                        self.wplat[self.iactwp],  self.wplon[self.iactwp],\
                        self.wplat[self.iactwp+1],self.wplon[self.iactwp+1])
        else:
            nextqdr = -999.
        return nextqdr
    
    @stack.command
    @staticmethod
    def cruisespd(acidx: 'acid', spd: 'spd'):
        """Sets the cruise speed of an aircraft in the autopilot. This speed is applied
        after the aircraft exits a turn.
        """
        
        # First, make sure that the velocity is within the performance limits of the aircraft.
        minspd = bs.traf.perf.vmin[acidx]
        maxspd = bs.traf.perf.vmax[acidx]
        spd_to_apply = min(maxspd, max(spd , minspd))
        # Apply the speed
        bs.traf.ap.cruisespd[acidx] = spd_to_apply
        return



''' Node test '''
import inspect
import zmq
import msgpack
import bluesky as bs
from bluesky import stack
from bluesky.core import Entity, Signal
from bluesky.network import context as ctx
from bluesky.network.subscriber import Subscription
from bluesky.network.npcodec import encode_ndarray, decode_ndarray
from bluesky.network.common import genid, asbytestr, seqidx2id, seqid2idx, MSG_SUBSCRIBE, MSG_UNSUBSCRIBE, GROUPID_NOGROUP, GROUPID_CLIENT, GROUPID_SIM, GROUPID_DEFAULT, IDLEN


# Register settings defaults
bs.settings.set_variable_defaults(recv_port=11000, send_port=11001)


class Node(Entity):
    def __init__(self, group_id=None):
        super().__init__()
        self.node_id = genid(group_id or GROUPID_NOGROUP)
        self.group_id = asbytestr(group_id or GROUPID_NOGROUP)[:len(self.node_id)-1]
        self.server_id = self.node_id[:-1] + seqidx2id(0)
        self.act_id = None
        self.nodes = set()
        self.servers = set()

        zmqctx = zmq.Context.instance()
        self.sock_recv = zmqctx.socket(zmq.SUB)
        self.sock_send = zmqctx.socket(zmq.XPUB)
        self.poller = zmq.Poller()

        # Tell bluesky that this client will manage the network I/O
        bs.net = self

        # Subscribe to subscriptions that were already made before constructing
        # this node
        for sub in Subscription.subscriptions.values():
            sub.subscribe_all()

        # Signals
        self.node_added = Signal('node-added')
        self.node_removed = Signal('node-removed')

    def connect(self, hostname=None, recv_port=None, send_port=None, protocol='tcp'):
        ''' Connect node to a server.

            Arguments:
            - hostname: Network name or ip of the server to connect to
            - recv_port: Network port to use for incoming communication
            - send_port: Network port to use for outgoing communication
            - protocol: Network protocol to use
        '''
        conbase = f'{protocol}://{hostname or "localhost"}'
        rcon = conbase + f':{recv_port or bs.settings.recv_port}'
        scon = conbase + f':{send_port or bs.settings.send_port}'
        self.sock_recv.connect(rcon)
        self.sock_send.connect(scon)
        self.poller.register(self.sock_recv, zmq.POLLIN)
        self.poller.register(self.sock_send, zmq.POLLIN)
        # Register this node by subscribing to targeted messages
        self.subscribe('',  '', to_group=self.node_id)

    def close(self):
        ''' Close all network connections. '''
        self.sock_recv.close()
        self.sock_send.close()
        zmq.Context.instance().destroy()


    def update(self):
        ''' Node periodic update function.

            Periodically call this function to allow this node to receive
            and process data.
        '''
        # Check for incoming data
        self.receive()

    def receive(self, timeout=0):
        ''' Poll for incoming data from Server, and receive if available.
            Arguments:
            timeout: The polling timeout in milliseconds. '''
        try:
            events = dict(self.poller.poll(timeout))
            # The socket with incoming data
            for sock, event in events.items():
                if event != zmq.POLLIN:
                    # The event does not refer to incoming data: skip for now
                    continue
            
                # Receive the message
                ctx.msg = sock.recv_multipart()
                if not ctx.msg:
                    # In the rare case that a message is empty, skip remaning processing
                    continue

                # Regular incoming data
                if sock == self.sock_recv:
                    ctx.topic = ctx.msg[0][IDLEN:-IDLEN].decode()
                    ctx.sender_id = ctx.msg[0][-IDLEN:]
                    pydata = msgpack.unpackb(ctx.msg[1], object_hook=decode_ndarray, raw=False)
                    sub = Subscription.subscriptions.get(ctx.topic, None) #or Subscription(ctx.topic, directedonly=True)
                    if sub is None:
                        print('No subscription known for', ctx.topic, 'on', self.node_id)
                        continue

                    # Unpack dict or list, skip empty string
                    if pydata == '':
                        sub.emit()
                    elif isinstance(pydata, dict):
                        sub.emit(**pydata)
                    elif isinstance(pydata, (list, tuple)):
                        sub.emit(*pydata)
                    else:
                        sub.emit(pydata)
                    ctx.msg = ctx.topic = ctx.sender_id = None

                elif sock == self.sock_send:
                    # This is an (un)subscribe message. If it's an id-only subscription
                    # this is also a registration message
                    if len(ctx.msg[0]) == IDLEN + 1:
                        sender_id = ctx.msg[0][1:]
                        sequence_idx = seqid2idx(sender_id[-1])
                        if sender_id[0] in (GROUPID_SIM, GROUPID_NOGROUP):
                            # This is an initial simulation node subscription
                            if ctx.msg[0][0] == MSG_SUBSCRIBE:
                                if sequence_idx > 0:
                                    self.nodes.add(sender_id)
                                    if sender_id != self.node_id:
                                        self.node_added.emit(sender_id)
                                    # if not self.act_id:
                                    #     self.actnode(sender_id)
                                    continue
                                elif sequence_idx == 0:
                                    self.servers.add(sender_id)

                            elif ctx.msg[0][0] == MSG_UNSUBSCRIBE:
                                if sequence_idx > 0:
                                    self.nodes.discard(sender_id)
                                    self.node_removed.emit(sender_id)
                                elif sequence_idx == 0:
                                    self.servers.discard(sender_id)

        except zmq.ZMQError:
            return False

    def send(self, topic, data='', to_group=''):
        btopic = asbytestr(topic)
        bto_group = asbytestr(to_group or stack.sender() or '')
        self.sock_send.send_multipart(
            [
                bto_group.ljust(IDLEN, b'*') + btopic + self.node_id,
                msgpack.packb(data, default=encode_ndarray, use_bin_type=True)
            ]
        )

    def subscribe(self, topic, from_group=GROUPID_DEFAULT, to_group=''):
        ''' Subscribe to a topic.

            Arguments:
            - topic: The name of the topic to subscribe to
            - from_id: The id of the node from which to receive the topic (optional)
            - to_group: The group mask that this topic is sent to (optional)
        '''
        sub = None
        if topic:
            sub = Subscription(topic)
            if (from_group, to_group) in sub.subs:
                # Subscription already active. Just return Subscription object
                return sub
            sub.subs.add((from_group, to_group))

        self._subscribe(topic, from_group, to_group)

        # Messages coming in that match this subscription will be emitted using a 
        # subscription signal
        return sub

    def unsubscribe(self, topic, from_group=GROUPID_DEFAULT, to_group=''):
        ''' Unsubscribe from a topic.

            Arguments:
            - topic: The name of the stream to unsubscribe from.
            - from_id: When subscribed to data from a specific node: The id of the node
            - to_group: The group mask that this topic is sent to (optional)
        '''
        if topic:
            Subscription(topic).subs.discard((from_group, to_group))
        self._unsubscribe(topic, from_group, to_group)

    def _subscribe(self, topic, from_group=GROUPID_DEFAULT, to_group=''):
        if from_group == GROUPID_DEFAULT:
            from_group = GROUPID_CLIENT
        btopic = asbytestr(topic)
        bfrom_group = asbytestr(from_group)
        bto_group = asbytestr(to_group)
        self.sock_recv.setsockopt(zmq.SUBSCRIBE, bto_group.ljust(IDLEN, b'*') + btopic + bfrom_group)

    def _unsubscribe(self, topic, from_group=GROUPID_DEFAULT, to_group=''):
        if from_group == GROUPID_DEFAULT:
            from_group = GROUPID_CLIENT
        btopic = asbytestr(topic)
        bfrom_group = asbytestr(from_group)
        bto_group = asbytestr(to_group)
        self.sock_recv.setsockopt(zmq.UNSUBSCRIBE, bto_group.ljust(IDLEN, b'*') + btopic + bfrom_group)

    def addnodes(self, count=1, *node_ids):
        ''' Tell the server to add 'count' nodes. 
        
            If provided, create these nodes with the specified node ids.
        '''
        self.send(b'ADDNODES', dict(count=count, node_ids=node_ids), self.server_id)


# Logic for group commands
import numpy as np
import bluesky as bs
from bluesky.core import TrafficArrays
from bluesky.tools import areafilter

class GroupArray(np.ndarray):
    # Similar to normal numpy arrays, but with the attribute of a groupname
    def __new__(cls,*args,groupname = '', **kwargs):
        ret = np.array(*args, **kwargs).view(cls)
        ret.groupname = groupname
        return ret
        
class TrafficGroups(TrafficArrays):
    def __init__(self):
        # Initialize the groups structure
        super().__init__()
        self.groups = dict()
        self.allmasks = 0
        with self.settrafarrays():
            self.ingroup = np.array([], dtype=np.int64)

    def __contains__(self, groupname):
        # Check if a group with a name exists
        return groupname in self.groups or groupname == '*'

    def group(self, groupname='', *args):
        '''Add aircraft to group, list aircraft in group, or list existing groups.'''
        # Return list of groups if no groupname is given
        if not groupname:
            if not self.groups:
                return True, 'There are currently no traffic groups defined.'
            else:
                return True, 'Defined traffic groups:\n' + ', '.join(self.groups)
        if len(self.groups) >= 64:
            return False, 'Maximum number of 64 groups reached'
        if groupname not in self.groups:
            if not args:
                return False, f'Group {groupname} doesn\'t exist'
            # Get first unused group mask
            for i in range(64):
                groupmask = (1 << i)
                if not self.allmasks & groupmask:
                    self.allmasks |= groupmask
                    self.groups[groupname] = groupmask
                    break

        elif not args:
            acnames = np.array(bs.traf.id)[self.listgroup(groupname)]
            return True, 'Aircraft in group {}:\n{}'.format(groupname, ', '.join(acnames))

        # Add aircraft to group
        if areafilter.hasArea(args[0]):
            inside = areafilter.checkInside(
                args[0], bs.traf.lat, bs.traf.lon, bs.traf.alt)
            self.ingroup[inside] |= self.groups[groupname]
            acnames = np.array(bs.traf.id)[inside]
        else:
            idx = list(args)
            self.ingroup[idx] |= self.groups[groupname]
            acnames = np.array(bs.traf.id)[idx]
        return True, 'Aircraft added to group {}:\n{}'.format(groupname, ', '.join(acnames))

    def delgroup(self, grouparray):
        ''' Delete a group, and all aircraft in that group. '''
        # Delete all aircraft in the respective group
        bs.traf.delete(grouparray)

        # Remove the group from the group list
        if grouparray.groupname != '*':
            self.allmasks ^= self.groups.pop(grouparray.groupname)

    def ungroup(self, groupname, *args):
        ''' Remove members from the group by aircraft id. '''
        groupmask = self.groups.get(groupname, None)
        if groupmask is None:
            return False, f"Group {groupname} doesn't exist"
        self.ingroup[list(args)] ^= groupmask

    def listgroup(self, groupname):
        ''' Return aircraft index for all aircraft in group. 
            When * is passed as groupname, all aircraft in simulation are returned. '''
        if groupname == '*':
            return GroupArray(range(bs.traf.ntraf), groupname='*')
        groupmask = self.groups.get(groupname, None)
        if groupmask is None:
            return False, f"Group {groupname} doesn't exist"
        return GroupArray(np.where((self.ingroup & groupmask) > 0)[0], groupname=groupname)



""" BlueSky deletion area plugin. This plugin can use an area definition to
    delete aircraft that exit the area. Statistics on these flights can be
    logged with the FLSTLOG logger. """
import numpy as np
# Import the global bluesky objects. Uncomment the ones you need
from bluesky import traf, sim  #, settings, navdb, traf, sim, scr, tools
from bluesky.tools import datalog, areafilter
from bluesky.core import Entity, timed_function
from bluesky.tools.aero import ft,kts,nm,fpm

# Log parameters for the flight statistics log
flstheader = \
    '#######################################################\n' + \
    'FLST LOG\n' + \
    'Flight Statistics\n' + \
    '#######################################################\n\n' + \
    'Parameters [Units]:\n' + \
    'Deletion Time [s], ' + \
    'Call sign [-], ' + \
    'Spawn Time [s], ' + \
    'Flight time [s], ' + \
    'Actual Distance 2D [nm], ' + \
    'Actual Distance 3D [nm], ' + \
    'Work Done [MJ], ' + \
    'Latitude [deg], ' + \
    'Longitude [deg], ' + \
    'Altitude [ft], ' + \
    'TAS [kts], ' + \
    'Vertical Speed [fpm], ' + \
    'Heading [deg], ' + \
    'Origin Lat [deg], ' + \
    'Origin Lon [deg], ' + \
    'Destination Lat [deg], ' + \
    'Destination Lon [deg], ' + \
    'ASAS Active [bool], ' + \
    'Pilot ALT [ft], ' + \
    'Pilot SPD (TAS) [kts], ' + \
    'Pilot HDG [deg], ' + \
    'Pilot VS [fpm]'  + '\n'

confheader = \
    '#######################################################\n' + \
    'CONF LOG\n' + \
    'Conflict Statistics\n' + \
    '#######################################################\n\n' + \
    'Parameters [Units]:\n' + \
    'Simulation time [s], ' + \
    'Total number of conflicts in exp area [-]\n'

# Global data
area = None

### Initialization function of your plugin. Do not change the name of this
### function, as it is the way BlueSky recognises this file as a plugin.
def init_plugin():

    # Addtional initilisation code
    global area
    area = Area()

    # Configuration parameters
    config = {
        # The name of your plugin
        'plugin_name':     'AREA',

        # The type of this plugin. For now, only simulation plugins are possible.
        'plugin_type':     'sim'
        }

    stackfunctions = {
        'AREA': [
            'AREA Shapename/OFF or AREA lat,lon,lat,lon,[top,bottom]',
            '[float/txt,float,float,float,alt,alt]',
            area.set_area,
            'Define deletion area (aircraft leaving area are deleted)'
        ],
        'EXP': [
            'EXP Shapename/OFF or EXP lat,lon,lat,lon,[top,bottom]',
            '[float/txt,float,float,float,alt,alt]',
            lambda *args: area.set_area(*args, exparea=True),
            'Define experiment area (area of interest)'
        ],
        'TAXI': [
            'TAXI ON/OFF [alt] : OFF auto deletes traffic below 1500 ft',
            'onoff[,alt]',
            area.set_taxi,
            'Switch on/off ground/low altitude mode, prevents auto-delete at 1500 ft'
        ]
    }
    # init_plugin() should always return these two dicts.
    return config, stackfunctions

class Area(Entity):
    ''' Traffic area: delete traffic when it leaves this area (so not when outside)'''
    def __init__(self):
        super().__init__()
        # Parameters of area
        self.active = False
        self.delarea = ''
        self.exparea = ''
        self.swtaxi = True  # Default ON: Doesn't do anything. See comments of set_taxi function below.
        self.swtaxialt = 1500.0  # Default alt for TAXI OFF
        self.prevconfpairs = set()
        self.confinside_all = 0

        # The FLST logger
        self.flst = datalog.crelog('FLSTLOG', None, flstheader)
        self.conflog = datalog.crelog('CONFLOG', None, confheader)

        with self.settrafarrays():
            self.insdel = np.array([], dtype=bool) # In deletion area or not
            self.insexp = np.array([], dtype=bool) # In experiment area or not
            self.oldalt = np.array([])
            self.distance2D = np.array([])
            self.distance3D = np.array([])
            self.dstart2D = np.array([])
            self.dstart3D = np.array([])
            self.workstart = np.array([])
            self.entrytime = np.array([])
            self.create_time = np.array([])

    def reset(self):
        ''' Reset area state when simulation is reset. '''
        super().reset()
        self.active = False
        self.delarea = ''
        self.exparea = ''
        self.swtaxi = True
        self.swtaxialt = 1500.0
        self.confinside_all = 0

    def create(self, n=1):
        ''' Create is called when new aircraft are created. '''
        super().create(n)
        self.oldalt[-n:] = traf.alt[-n:]
        self.insdel[-n:] = False
        self.insexp[-n:] = False
        self.create_time[-n:] = sim.simt

    @timed_function(name='AREA', dt=1.0)
    def update(self, dt):
        ''' Update flight efficiency metrics
            2D and 3D distance [m], and work done (force*distance) [J] '''
        if self.active:
            resultantspd = np.sqrt(traf.gs * traf.gs + traf.vs * traf.vs)
            self.distance2D += dt * traf.gs
            self.distance3D += dt * resultantspd

            # Find out which aircraft are currently inside the experiment area, and
            # determine which aircraft need to be deleted.
            insdel = areafilter.checkInside(self.delarea, traf.lat, traf.lon, traf.alt)
            insexp = insdel if not self.exparea else \
                areafilter.checkInside(self.exparea, traf.lat, traf.lon, traf.alt)
            # Find all aircraft that were inside in the previous timestep, but no
            # longer are in the current timestep
            delidx = np.where(np.array(self.insdel) * (np.array(insdel) == False))[0]
            self.insdel = insdel

            # Count new conflicts where at least one of the aircraft is inside
            # the experiment area
            # Store statistics for all new conflict pairs
            # Conflict pairs detected in the current timestep that were not yet
            # present in the previous timestep
            confpairs_new = list(set(traf.cd.confpairs) - self.prevconfpairs)
            if confpairs_new:
                # If necessary: select conflict geometry parameters for new conflicts
                # idxdict = dict((v, i) for i, v in enumerate(traf.cd.confpairs))
                # idxnew = [idxdict.get(i) for i in confpairs_new]
                # dcpa_new = np.asarray(traf.cd.dcpa)[idxnew]
                # tcpa_new = np.asarray(traf.cd.tcpa)[idxnew]
                # tLOS_new = np.asarray(traf.cd.tLOS)[idxnew]
                # qdr_new = np.asarray(traf.cd.qdr)[idxnew]
                # dist_new = np.asarray(traf.cd.dist)[idxnew]

                newconf_unique = {frozenset(pair) for pair in confpairs_new}
                ac1, ac2 = zip(*newconf_unique)
                idx1 = traf.id2idx(ac1)
                idx2 = traf.id2idx(ac2)
                newconf_inside = np.logical_or(insexp[idx1], insexp[idx2])

                nnewconf_exp = np.count_nonzero(newconf_inside)
                if nnewconf_exp:
                    self.confinside_all += nnewconf_exp
                    self.conflog.log(self.confinside_all)
            self.prevconfpairs = set(traf.cd.confpairs)

            # Register distance values upon entry of experiment area
            newentries = np.logical_not(self.insexp) * insexp
            self.dstart2D[newentries] = self.distance2D[newentries]
            self.dstart3D[newentries] = self.distance3D[newentries]
            self.workstart[newentries] = traf.work[newentries]
            self.entrytime[newentries] = sim.simt

            # Log flight statistics when exiting experiment area
            exits = np.logical_and(self.insexp,np.logical_not(insexp))
            # Update insexp
            self.insexp = insexp

            if np.any(exits):
                self.flst.log(
                    np.array(traf.id)[exits],
                    self.create_time[exits],
                    sim.simt - self.entrytime[exits],
                    (self.distance2D[exits] - self.dstart2D[exits])/nm,
                    (self.distance3D[exits] - self.dstart3D[exits])/nm,
                    (traf.work[exits] - self.workstart[exits])*1e-6,
                    traf.lat[exits],
                    traf.lon[exits],
                    traf.alt[exits]/ft,
                    traf.tas[exits]/kts,
                    traf.vs[exits]/fpm,
                    traf.hdg[exits],
                    traf.cr.active[exits],
                    traf.aporasas.alt[exits]/ft,
                    traf.aporasas.tas[exits]/kts,
                    traf.aporasas.vs[exits]/fpm,
                    traf.aporasas.hdg[exits])

            # delete all aicraft in self.delidx
            if len(delidx) > 0:
                traf.delete(delidx)



        # Autodelete for descending with swTaxi:
        if not self.swtaxi:
            delidxalt = np.where((self.oldalt >= self.swtaxialt)
                                 * (traf.alt < self.swtaxialt))[0]
            self.oldalt = traf.alt
            if len(delidxalt) > 0:
                traf.delete(list(delidxalt))

    def set_area(self, *args, exparea=False):
        ''' Set Experiment Area. Aircraft leaving the experiment area are deleted.
        Input can be existing shape name, or a box with optional altitude constraints.'''
        curname = self.exparea if exparea else self.delarea
        msgname = 'Experiment area' if exparea else 'Deletion area'
        # if all args are empty, then print out the current area status
        if not args:
            return True, f'{msgname} is currently ON (name={curname})' if self.active else \
                         f'{msgname} is currently OFF'

        # start by checking if the first argument is a string -> then it is an area name
        if isinstance(args[0], str) and len(args)==1:
            if areafilter.hasArea(args[0]):
                # switch on Area, set it to the shape name
                if exparea:
                    self.exparea = args[0]
                else:
                    self.delarea = args[0]

                self.active = True
                self.flst.start()
                self.conflog.start()
                return True, f'{msgname} is set to {args[0]}'
            if args[0][:2] =='OF':
                # switch off the area and reset the logger
                self.active = False
                return True, f'{msgname} is switched OFF'
            if args[0][:2] == 'ON':
                if not self.name:
                    return False, 'No area defined.'
                else:
                    self.active = True
                    return True, f'{msgname} switched ON (name={curname})'
            # shape name is unknown
            return False, 'Shapename unknown. ' + \
                'Please create shapename first or shapename is misspelled!'
        # if first argument is a float -> then make a box with the arguments
        if isinstance(args[0],(float, int)) and 4<=len(args)<=6:
            self.active = True
            if exparea:
                self.exparea = 'EXPAREA'
                areafilter.defineArea('EXPAREA', 'BOX', args[:4], *args[4:])
            else:
                self.delarea = 'DELAREA'
                areafilter.defineArea('DELAREA', 'BOX', args[:4], *args[4:])
            self.flst.start()
            self.conflog.start()
            return True, f'{msgname} is ON. Area name is: {"EXP" if exparea else "DEL"}AREA'

        return False,  'Incorrect arguments' + \
                       '\nAREA Shapename/OFF or\n Area lat,lon,lat,lon,[top,bottom]'

    def set_taxi(self, flag,alt=1500*ft):
        ''' Taxi ON/OFF to autodelete below a certain altitude if taxi is off'''
        self.swtaxi = flag # True =  taxi allowed, False = autodelete below swtaxialt
        self.swtaxialt = alt



''' Simulation clock with guaranteed decimal precision. '''
from types import SimpleNamespace
from decimal import Decimal

from bluesky import settings


# Register settings defaults
settings.set_variable_defaults(simdt=0.05)

MAX_RECOVERY_FAC = 4

# Data that the simulation clock needs to keep
_clock = SimpleNamespace(t=Decimal('0.0'), dt=Decimal(repr(settings.simdt)),
                         ft=0.0, fdt=settings.simdt)


def setdt(newdt=None, target='simdt'):
    ''' Set the timestep for the simulation clock.
        Returns a floating-point representation of the new timestep. '''
    if newdt is None:
        text = f'Simulation timesteps:\nbase dt = {_clock.fdt}'
        for timer in Timer.timers():
            text += f'\n{timer.name} = {timer.dt_act}'
        return True, text
    if target == 'simdt':
        _clock.dt = Decimal(repr(newdt))
        _clock.fdt = float(_clock.dt)
        msg = f'Base dt set to {_clock.dt}'
        for timer in Timer.timers():
            _, tmsg = timer.setdt()
            msg = msg + '\n' + tmsg

        return True, msg
    timer = Timer.gettimer(target)
    if timer is None:
        return False, f'Timer {target} not found'
    return timer.setdt(newdt)


def step(recovery_time=0):
    ''' Increment the time of this clock with one timestep, plus a possible
        recovery time increment if the simulation is lagging and real-time
        running is enabled.
        Returns a floating-point representation of the new simulation time,
        and the actual timestep. '''
    recovery_time = min(Decimal(recovery_time), MAX_RECOVERY_FAC * _clock.dt)
    _clock.t += _clock.dt + recovery_time
    _clock.ft = float(_clock.t)
    for timer in Timer.timers():
        timer.step()

    return _clock.ft, _clock.fdt + float(recovery_time)


def reset():
    ''' Reset function executed when simulation is reset.'''
    # Reset the simulation clock.
    _clock.t = Decimal('0.0')
    _clock.dt = Decimal(repr(settings.simdt))
    _clock.ft = 0.0
    _clock.fdt = float(_clock.dt)
    for timer in Timer.timers():
        timer.reset()


class TimerMeta(type):
    __timers__ = dict()

    def __call__(cls, name, dt):
        timer = TimerMeta.__timers__.get(name)
        if timer is None:
            TimerMeta.__timers__[name] = timer = super().__call__(name, dt)
        return timer

    @classmethod
    def timers(cls):
        return cls.__timers__.values()

    @classmethod
    def gettimer(cls, name):
        return cls.__timers__.get(name)


class Timer(metaclass=TimerMeta):
    ''' Timer class for simulation-time periodic functions. '''

    def __init__(self, name, dt):
        self.name = name
        self.dt_default = Decimal(repr(dt))
        self.dt_requested = self.dt_default
        self.dt_act = self.dt_default
        self.rel_freq = 0
        self.counter = 0
        self.tprev = _clock.t
        self.setdt()

    def reset(self):
        ''' Reset all simulation timers to their default time interval. '''
        self.dt_requested = self.dt_default
        self.dt_act = self.dt_default
        self.rel_freq = 0
        self.counter = 0
        self.tprev = _clock.t
        self.setdt()

    def setdt(self, dt=None):
        ''' Set the update interval of this timer. '''
        # setdt is called without arguments if the base dt has changed
        # In this case, check if our dt is still ok.
        if dt:
            # Store the requested dt separately: is used to update actual dt
            # for when the simulation dt is changed
            self.dt_requested = Decimal(repr(dt))

        # Calculate the relative frequency of the simulation with respect to this timer
        rel_freq = max(1, int(self.dt_requested // _clock.dt))
        # Update timer to next trigger point
        passed = self.rel_freq - self.counter
        self.counter = max(0, rel_freq - passed)
        self.rel_freq = rel_freq
        dtnew = self.rel_freq * _clock.dt
        if abs(self.dt_act - dtnew) < 0.0001:
            return True, self.name + ' dt is unchanged.'
        self.dt_act = dtnew
        if abs(self.dt_act - self.dt_requested) > 0.0001:
            return True, self.name + \
                f' dt set to {self.dt_act} to match integer multiple of base dt.'
        return True, self.name + f' dt set to {self.dt_act}'

    def step(self):
        ''' Step is called each base timestep to update this timer. '''
        self.counter = (self.counter or self.rel_freq) - 1

    def readynext(self):
        ''' Returns True if a time interval of this timer has passed. '''
        return self.counter == 0

    def elapsed(self):
        ''' Return the time elapsed since the last time this timer was triggered. '''
        elapsed = float(_clock.t - self.tprev)
        self.tprev = _clock.t
        return elapsed



from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QSplitter, QSplitterHandle, QVBoxLayout, QToolButton


class Handle(QSplitterHandle):
    ''' Custom splitter handle with collapse button. '''
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.vl = QVBoxLayout(self)
        self.vl.setContentsMargins(0, 0, 0, 0)
        self.b = QToolButton(self)
        self.b.setArrowType(Qt.ArrowType.LeftArrow)
        self.b.clicked.connect(self.parent().collapse)
        self.vl.addWidget(self.b)
        self.setLayout(self.vl)


class Splitter(QSplitter):
    ''' Custom splitter with collapse button.
    
        A Splitter divides resizeable panes within a window. 
    '''
    def __init__(self, *args, defwidth=200, **kwargs):
        super().__init__(*args, **kwargs)
        self.prevstate = [10000, defwidth]
        self.splitterMoved.connect(self.on_splitter_move)

    def createHandle(self):
        return Handle(self.orientation(), self)

    def collapse(self):
        ''' Callback for the collapse button on the Splitter handle. '''
        sizes = self.sizes()
        if sizes[-1] == 0:
            sizes = self.prevstate
            self.handle(1).b.setArrowType(Qt.ArrowType.RightArrow)

        else:
            self.prevstate = sizes[:]
            sizes[-1] = 0
            self.handle(1).b.setArrowType(Qt.ArrowType.LeftArrow)
        self.setSizes(sizes)

    def on_splitter_move(self, pos, index):
        ''' Keep track of Splitter position to update handle appearance. '''
        handle = self.handle(1)
        if self.width() - handle.width() - pos == 0:
            handle.b.setArrowType(Qt.ArrowType.LeftArrow)
        else:
            handle.b.setArrowType(Qt.ArrowType.RightArrow)



""" BlueSky aircraft performance calculations."""
from xml.etree import ElementTree
from math import *
import numpy as np
from bluesky.tools.aero import ft, g0, rho0, kts, lbs, inch, sqft, fpm

from .performance import esf, phases, calclimits, PHASE
import bluesky as bs

# Register settings defaults
bs.settings.set_variable_defaults(perf_path='performance/BS', verbose=False)

class CoeffBS:
    """
    Coefficient class definition : get aircraft-specific coefficients from database
    Created by  : Isabel Metz

    References:

    - D.P. Raymer. Aircraft Design: A Conceptual Approach. AIAA Education Series.
    American Institute of Aeronautics and Astronautics, Inc., Reston, U.S, fifth edition, 2012.
    - R. Babikian. The Historical Fuel Efficiency Characteristics of Regional Aircraft from
    Technological, Operational, and Cost Perspectives. Master's Thesis, Massachusetts
    Institute of Technology, Boston, U.S.
    """

    def convert(self, value, unit):
        factors = {'kg': 1., 't':1000., 'lbs': lbs, 'N': 1., 'W': 1, \
                    'm':1.,'km': 1000., 'inch': inch,'ft': ft, \
                    'sqm': 1., 'sqft': sqft, 'sqin': 0.0254*0.0254 ,\
                    'm/s': 1., 'km/h': 1./3.6, 'kts': kts, 'fpm': fpm, \
                    "kg/s": 1., "kg/m": 1./60., 'mug/J': 0.000001, 'mg/J': 0.001 ,
                    "kW": 1000.,"kN":1000.,
                    "":1.}

        if unit in factors:
            converted = factors[unit] * float(value)

        else:
            converted = float(value)
            if not self.warned:
                print("traf/perf.py convert function: Unit mismatch. Could not find ", unit)
                self.warned = True

        return converted


    def coeff(self):

        # aircraft
        self.atype     = [] # aircraft type
        self.j_ac      = [] # list of all jet aircraft
        self.tp_ac     = [] # list of all turboprop aircraft

        # engine
        self.etype     = [] # jet / turboprop
        self.engines   = [] # engine types avaliable per aircraft type
        self.j_engines = [] # engine types for jet aircraft
        self.tp_engines= [] # engine types for turboprop aircraft
        self.n_eng     = [] # number of engines

        # weights
        self.MTOW      = [] # maximum takeoff weight

        # speeds
        self.max_spd   = [] # maximum CAS
        self.cr_Ma     = [] # nominal cruise Mach at 35000 ft
        self.cr_spd    = [] # cruise speed
        self.max_Ma    = [] # maximum Mach
        self.gr_acc    = [] # ground acceleration
        self.gr_dec    = [] # ground deceleration

        # limits
        self.vmto      = [] # minimum speed during takeoff
        self.vmld      = [] # minimum speed during landing
        self.clmax_cr  = [] # max. cruise lift coefficient
        self.max_alt   = [] # maximum altitude

        # dimensions
        #span          = [] # wing span
        self.Sref      = [] # reference wing area
        #wet_area      = [] # wetted area

        # aerodynamics
        #Cfe           = [] # equivalent skin friction coefficient (Raymer, p.428)
        self.CD0       = [] # parasite drag coefficient
        #oswald        = [] # oswald factor
        self.k         = [] # induced drag factor

        # scaling factors for drag (FAA_2005 SAGE)
        # order of flight phases: TO, IC, CR ,AP, LD ,LD gear
        self.d_CD0j    = [1.476, 1.143,1.0, 1.957, 3.601, 1.037]
        self.d_kj      = [1.01, 1.071, 1.0 ,0.992, 0.932, 1.0]
        self.d_CD0t    = [1.220, 1.0, 1.0, 1.279, 1.828, 0.496]
        self.d_kt      = [0.948, 1.0, 1.0, 0.94, 0.916, 1.0]

        # bank angles per phase. Order: TO, IC, CR, AP, LD. Currently already in CTraffic
        # self.bank = np.deg2rad(np.array([15,35,35,35,15]))

        # flag: did we already warn about invalid input unit?
        self.warned    = False

        # parse AC files

        path = bs.resource(bs.settings.perf_path) / 'BS/aircraft'
        for fname in path.iterdir():
            acdoc = ElementTree.parse(fname)

            #actype = doc.find('ac_type')
            self.atype.append(acdoc.find('ac_type').text)

            # engine
            self.etype.append(int(acdoc.find('engine/eng_type').text))

            # store jet and turboprop aircraft in seperate lists for accessing specific engine data
            if int(acdoc.find('engine/eng_type').text) ==1:
                self.j_ac.append(acdoc.find('ac_type').text)

            elif int(acdoc.find('engine/eng_type').text) ==2:
                self.tp_ac.append(acdoc.find('ac_type').text)

            self.n_eng.append(float(acdoc.find('engine/num_eng').text))

            engine = []
            for eng in acdoc.findall('engine/eng'):
                engine.append(eng.text)

            # weights
            MTOW = self.convert(acdoc.find('weights/MTOW').text, acdoc.find('weights/MTOW').attrib['unit'])

            self.MTOW.append(MTOW)

            MLW= self.convert(acdoc.find('weights/MLW').text, acdoc.find('weights/MLW').attrib['unit'])

            # dimensions
            # wingspan
            span = self.convert(acdoc.find('dimensions/span').text, acdoc.find('dimensions/span').attrib['unit'])
            # reference surface area
            S_ref = self.convert(acdoc.find('dimensions/wing_area').text, acdoc.find('dimensions/wing_area').attrib['unit'])
            self.Sref.append(S_ref)

            # wetted area
            S_wet = self.convert(acdoc.find('dimensions/wetted_area').text, acdoc.find('dimensions/wetted_area').attrib['unit'])

            # speeds
            # cruise Mach number
            crma = acdoc.find('speeds/cr_MA')
            if float(crma.text) == 0.0:
                # to be refined
                self.cr_Ma.append(0.8)
            else:
                self.cr_Ma.append(float(crma.text))

            # cruise TAS
            crspd = acdoc.find('speeds/cr_spd')

            # to be refined
            if float(crspd.text) == 0.0:
                self.cr_spd.append(self.convert(250, 'kts'))
            else:
                self.cr_spd.append(self.convert(acdoc.find('speeds/cr_spd').text, acdoc.find('speeds/cr_spd').attrib['unit']))

            # ground acceleration
            # values are based on statistical ADS-B evaluations
            # turboprops: 2.12 m/s^2 acceleration,1.12m/s^2 deceleration
            if int(acdoc.find('engine/eng_type').text) == 2:
                self.gr_acc.append(2.12)
                self.gr_dec.append(1.12)

            # turbofans
            else:

                # turbofans with two engines: 1.94 m/^2, 1.265m/s^2 deceleration
                if float(acdoc.find('engine/num_eng').text) == 2. :
                    self.gr_acc.append(1.94)
                    self.gr_dec.append(1.265)
                # turbofans with four engines: 1.68 m/s^2, 1.131 m/s^2 deceleration
                #  assumption: aircraft with three engines have the same value
                else :
                    self.gr_acc.append(1.68)
                    self.gr_dec.append(1.131)



            # limits
            # min takeoff speed
            tospd = acdoc.find('speeds/to_spd')
            # no take-off speed given: calculate via cl_max
            if float (tospd.text) == 0.:
                clmax_to = float(acdoc.find('aerodynamics/clmax_to').text)
                self.vmto.append (sqrt((2*g0)/(S_ref*clmax_to))) # influence of current weight and density follows in CTraffic
            else:
                tospd = self.convert(acdoc.find('speeds/to_spd').text, acdoc.find('speeds/to_spd').attrib['unit'])
                self.vmto.append(tospd/(1.13*sqrt(MTOW/rho0))) # min spd according to CS-/FAR-25.107
            # min ic, cr, ap speed
            clmaxcr = (acdoc.find('aerodynamics/clmax_cr'))
            self.clmax_cr.append(float(clmaxcr.text))

            # min landing speed
            ldspd = acdoc.find('speeds/ld_spd')
            if float(ldspd.text) == 0. :
                clmax_ld = (acdoc.find('aerodynamics/clmax_ld'))
                self.vmld.append (sqrt((2*g0)/(S_ref*float(clmax_ld.text)))) # influence of current weight and density follows in CTraffic
            else:
                ldspd = self.convert(acdoc.find('speeds/ld_spd').text, acdoc.find('speeds/ld_spd').attrib['unit'])
                clmax_ld = MLW*g0*2/(rho0*(ldspd*ldspd)*S_ref)
                self.vmld.append(ldspd/(1.23*sqrt(MLW/rho0)))
            # maximum CAS
            maxspd = acdoc.find('limits/max_spd')
            if float(maxspd.text) == 0.0:
                # to be refined
                self.max_spd.append(400.)
            else:
                self.max_spd.append(self.convert(acdoc.find('limits/max_spd').text, acdoc.find('limits/max_spd').attrib['unit']))
            # maximum Mach
            maxma = acdoc.find('limits/max_MA')
            if float(maxma.text) == 0.0:
                # to be refined
                self.max_Ma.append(0.8)
            else:
                self.max_Ma.append(float(maxma.text))


            # maximum altitude
            maxalt = acdoc.find('limits/max_alt')
            if float(maxalt.text) == 0.0:
                #to be refined
                self.max_alt.append(11000.)
            else:
                self.max_alt.append(self.convert(acdoc.find('limits/max_alt').text, acdoc.find('limits/max_alt').attrib['unit']))

            # aerodynamics

            # parasitic drag - according to Raymer, p. 429
            Cfe = float((acdoc.find('aerodynamics/Cfe').text))
            self.CD0.append (Cfe*S_wet/S_ref)

            # induced drag
            oswald = acdoc.find('aerodynamics/oswald')
            if float(oswald.text) == 0.0:
                # math method according to Obert 2009, p.542: e = 1/(1.02+0.09*pi*AR) combined with Nita 2012, p.2
                self.k.append(1.02/(pi*(span*span/S_ref))+0.009)
            else:
                oswald = float(acdoc.find('aerodynamics/oswald').text)
                self.k.append(1/(pi*oswald*(span*span/S_ref)))

            #users = doc.find( 'engine' )
            #for node in users.getiterator():
            #    print node.tag, node.attrib, node.text, node.tail

            # to collect avaliable engine types per aircraft
            # 2do!!! access via console so user may choose preferred engine
            # for data file: statistics provided by flightglobal for first choice
            # if not declared differently: first engine is taken!
            self.engines.append(engine)

            if int(acdoc.find('engine/eng_type').text) ==1:
                self.j_engines.append(engine)

            elif int(acdoc.find('engine/eng_type').text) ==2:
                self.tp_engines.append(engine)

        # engines
        self.enlist      = [] # list of all engines
        self.jetenlist   = [] # list of all jet engines
        self.propenlist  = [] # list of all turbopropengines

        # a. jet aircraft
        self.rated_thrust = [] # rated Thrust (one engine)
        self.ffto        = [] # fuel flow takeoff
        self.ffcl        = [] # fuel flow climb
        self.ffcr        = [] # fuel flow cruise
        self.ffid        = [] # fuel flow idle
        self.ffap        = [] # fuel flow approach
        self.SFC         = [] # specific fuel flow cruise


        # b. turboprops
        self.P           = [] # max. power (Turboprops, one engine)
        self.PSFC_TO     = [] # SFC takeoff
        self.PSFC_CR     = [] # SFC cruise

        # parse engine files
        path = bs.resource(bs.settings.perf_path) / 'BS/engines/'
        for fname in path.iterdir():
            endoc = ElementTree.parse(fname)
            self.enlist.append(endoc.find('engines/engine').text)

            # thrust
            # a. jet engines
            if int(endoc.find('engines/eng_type').text) ==1:

                # store engine in jet-engine list
                self.jetenlist.append(endoc.find('engines/engine').text)
                # thrust
                self.rated_thrust.append(self.convert(endoc.find('engines/Thr').text, endoc.find('engines/Thr').attrib['unit']))
                # bypass ratio
                BPRc = int(endoc.find('engines/BPR_cat').text)
                # different SFC for different bypass ratios (reference: Raymer, p.36)
                SFC = [14.1, 22.7, 25.5]
                self.SFC.append(SFC[BPRc])

                # fuel flow: Takeoff, climb, cruise, approach, idle
                self.ffto.append(self.convert(endoc.find('ff/ff_to').text, endoc.find('ff/ff_to').attrib['unit']))
                self.ffcl.append(self.convert(endoc.find('ff/ff_cl').text, endoc.find('ff/ff_cl').attrib['unit']))
                self.ffcr.append(self.convert(endoc.find('ff/ff_cr').text, endoc.find('ff/ff_cr').attrib['unit']))
                self.ffap.append(self.convert(endoc.find('ff/ff_ap').text, endoc.find('ff/ff_ap').attrib['unit']))
                self.ffid.append(self.convert(endoc.find('ff/ff_id').text, endoc.find('ff/ff_id').attrib['unit']))

            # b. turboprop engines
            elif int(endoc.find('engines/eng_type').text) ==2:

                # store engine in prop-engine list
                self.propenlist.append(endoc.find('engines/engine').text)

                # power
                self.P.append(self.convert(endoc.find('engines/Power').text, endoc.find('engines/Power').attrib['unit']))
                # specific fuel consumption: takeoff and cruise
                PSFC_TO = self.convert(endoc.find('SFC/SFC_TO').text, endoc.find('SFC/SFC_TO').attrib['unit'])
                self.PSFC_TO.append(PSFC_TO)
                # according to Babikian (function based on PSFC in [mug/J]), input in [kg/J]
                self.PSFC_CR.append(self.convert((0.7675*PSFC_TO*1000000.0 + 23.576), 'mug/J'))
                # print PSFC_TO, self.PSFC_CR

        # Turn relevant ones into numpy arrays

        self.MTOW=np.array(self.MTOW)
        self.Sref=np.array(self.Sref)
        self.etype=np.array(self.etype)
        self.cr_Ma=np.array(self.cr_Ma)
        self.cr_spd=np.array(self.cr_spd)
        self.gr_acc=np.array(self.gr_acc)
        self.gr_dec=np.array(self.gr_dec)
        self.vmto=np.array(self.vmto)
        self.vmld=np.array(self.vmld)
        self.max_Ma=np.array(self.max_Ma)
        self.max_spd=np.array(self.max_spd)
        self.max_alt=np.array(self.max_alt)
        self.CD0=np.array(self.CD0)
        self.k=np.array(self.k)
        self.clmax_cr=np.array(self.clmax_cr)
        self.n_eng=np.array(self.n_eng)
        self.P=np.array(self.P)
        self.PSFC_TO=np.array(self.PSFC_TO)
        self.PSFC_CR=np.array(self.PSFC_CR)
        self.rated_thrust=np.array(self.rated_thrust)
        self.SFC=np.array(self.SFC)
        self.ffto=np.array(self.ffto)
        self.ffcl=np.array(self.ffcl)
        self.ffcr=np.array(self.ffcr)
        self.ffid=np.array(self.ffid)
        self.ffap=np.array(self.ffap)
